!+
! Module: asmooth_mod
!
! Authors:  Ian Stewart - SSC-LUX
!          
! Description: Performs adaptive smoothing of FITS images.
! 
! Comments:
!
!TODO:
!	- doco upgrade.
!	- check that, if min/maxsigma or ngauss are different to those used to make a template set, that the output still looks ok.
!	- Check accuracy of spher2ims for small (decA - decRef). (see equivalent rgslib routine)
!	- Do some careful conversion of real64 inputs.
!	- make sure test sets of other tasks fail if the task is not there! See GV mail of 29 May 2001. This is now done for asmooth.
!	- reinstitute the supercritical smoothing option?
!	- see if can reduce memory loading further by absorbing mask into indexImage.
!	- see if can speed up fft by sampling a single gaussian array instead of having to recalculate it all the time.
!	- Clean up code.
!	- 'samesnr' to 'sameamp' (see doco)
!	- change det mask to ccd mask??
!	- copy over keywords and dss stuff _before_ smoothing, because this is quicker than the smoothing but at the same time more likely to result in errors.
!	- Output if desired the smoothed variance (or the snr of the smoothed image?) (see SPR 2829, discussed this with JBB, who recommended it).
!	- Write more keywords to output images (to allow the user to identify the input dataset).
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module asmooth_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling
  use cal
  use caloaldefs
!use temp_ssclib
!use test_mod
  use datafilter

  use constants
  use fftw_aux
  use my_psf
  use coordinate
  use source_cutouts
  use linear
  use cal_aux
  use wcs_aux
  use fits_utils
  use array_utils
use dump_mod

  implicit none 

  private 

  ! Command-line parameters:
  logical ::&
    withInMask,&
    withOutMask,&
    writeBadMask,&
    writeCheeseMask,&
    readTemplate,&
    writeTemplate,&
    withSrcList,&
    withExpMap,&
    remultiplyByExpMap,&
    forceCalcType,&
    doCalcByFFT,&
    copyDss,&
    withVarianceSet
  character(len=255) ::&
    inSetName,&
    inMaskSetName,&
    outMaskSetName,&
    badMaskSetName,&
    cheeseMaskSetName,&
    templateSetName,&
    srcListSetTabName,&
    expMapSetName,&
    outSetName,&
    varianceSetName,&
    selectionExpression
  character(len=20) ::&
    raColName,&
    decColName,&
    rateColName,&
    bkgRateColName
  character(len=8) ::&
    smoothType,&
    expMapUseType
  real(single) ::&
    minGaussianSigma,&
    maxGaussianSigma,&
    desiredSNR,&
    gaussianSigma,&
    limitOnNumDirectOps,&
    fracBackground
  integer(int16) ::&
    numGaussians

  ! Other globals:
  integer(int16) ::&
    xStart,&
    yStart,&
    xFinis,&
    yFinis,&
    xsize,&
    ysize
  type(wcsDataType) ::&
    wcs
  integer(int32) ::&
    instrumentId
!,&
!    imageBandNum
!  character(len=8) ::&
!    wcsType
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude
  real(double) ::&
!    fovRadiusDeg,&
    expStartTimeStamp
!,&
!    fovCtrXworld,&
!    fovCtrYworld
  logical ::&
!    imageBandKwdExists,&
    forwardPlanExists = .false.,&
    reversePlanExists = .false.
  real(single), parameter ::&
    detUnitArcsec = 0.05,& ! The unit of the DETX,Y system.
!*** should be in constants module?? Or in a param file??
    minAllowedWeight = 0.0001
!  integer(int16) ::&
!    bandNum
  real(single), dimension(8), parameter ::&
    bandLoEeV = (/200.0, 500.0, 2000.0, 4500.0, 7500.0, 200.0, 2000.0, 200.0/)
  real(single), dimension(8), parameter ::&
    bandHiEeV = (/500.0, 2000.0, 4500.0, 7500.0, 12000.0, 2000.0, 12000.0, 12000.0/)
!***** These numbers (1-5,8 anyway) are from the products. Note however that the icd says band 8 should start at 100 eV.

  real(single) ::&
    maxSnrImprovement
  type gaussData
    real(single)   :: sigma, norm, snrImprovement
    integer(int32) :: numPixels
    integer(int16) :: halfWidthPixels
  end type gaussData
  type(FftwPlanType) :: fftForwardPlanNum, fftReversePlanNum
    ! type def is in fftw_aux.

integer :: txi = 331, tyi = 350

  public asmooth

contains 
  
  !---------------------------------------------------------------------
  subroutine asmooth
  !---------------------------------------------------------------------

    implicit none

    real(single), dimension(:,:), pointer ::&
      image
    integer(int8), dimension(:,:), pointer ::&
      mask
    real(single), dimension(:,:), allocatable ::&
      smoothedImage
integer :: minXi, maxXi, minYi, maxYi
character(8) :: numBadPixStr

    !-------------------------------------------------------------------
    ! Start

    call readParameters

!    if (withDetMask) then
!      call loadDetMask(mask) ! detMask > 0 only within detector FOV.
!      call loadInImage(inImage)
!    else
!      call loadInImage(inImage)
!      allocate(mask(xStart:xFinis, yStart:yFinis))
!      mask = 1
!    end if

call loadFitsArray(inSetName, 0, image)
xsize = size(image, 1)
ysize = size(image, 2)

call loadMask(xsize, ysize, mask)

call maskArrayLimits((mask > 0), minXi, maxXi, minYi, maxYi)
xStart = minXi
yStart = minYi
xFinis = maxXi
yFinis = maxYi
call selectImageSubset(image, xStart, xFinis, yStart, yFinis)
call selectImageSubset(mask,  xStart, xFinis, yStart, yFinis)

    allocate(smoothedImage(xStart:xFinis, yStart:yFinis))
!**** smoothedImage = 0.0 here????

!    call doSomeChecks

    if (withSrcList) then
      call addSrcMask(mask)  ! After this operation, (mask & 1) \= 0
      ! near sources.
    end if

    where (.not. btest(mask, 0)) image = 0.0 ! ie, .not. inMask
!*** is this necessary??

!........................................................................
    if (smoothType .eq. 'simple') then

      call simpleSmooth(image, mask, gaussianSigma, smoothedImage)

!........................................................................
    else ! smoothType .eq. 'adaptive'

      call adaptiveSmooth(image, mask, smoothedImage)

    end if ! smoothType .eq. 'simple'

!........................................................................

    if (any(btest(mask, 2))) then
      numBadPixStr = intToStr(count(btest(mask, 2)))
      write(errstr, '(a, f9.7, a)') 'The sum over weights was found to &
      &be smaller than the limit of ', minAllowedWeight&
      , ' at '//trim(numBadPixStr)//' pixels. The output has been set &
      &to zero at these pixels. A map of these pixels may be obtained &
      &by setting --writebadmaskset=yes.'

      call warning('outMaskTooBroad')

      where(btest(mask, 2)) smoothedImage = 0.0
    end if

    call message(AppMsg, Verbose, 'Writing smoothed output image '&
    //trim(outSetName)//'.')

    call saveOutImage(smoothedImage, inSetName, outSetName&
      , xsize, ysize, xStart, yStart)

    if (writeBadMask) then
      call message(AppMsg, Verbose, 'Writing bad pixel mask '&
      //trim(badMaskSetName)//'.')

      call saveMask(btest(mask, 2), badMaskSetName&
      , xsize, ysize, xStart, yStart)
    end if

    if (withSrcList .and. writeCheeseMask) then
      call message(AppMsg, Verbose, 'Writing cheese mask '&
      //trim(cheeseMaskSetName)//'.')

      call saveMask(btest(mask, 0), cheeseMaskSetName&
      , xsize, ysize, xStart, yStart)
    end if

    deallocate(smoothedImage, image, mask)

  end subroutine asmooth

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    inSetName = stringParameter('inset')
    outSetName = stringParameter('outset')
    copyDss = booleanParameter('copydss')

    smoothType = stringParameter('smoothtype')
    gaussianSigma = realParameter('sigma')
    numGaussians = intParameter('ngauss')
    minGaussianSigma = realParameter('minsigma')
    maxGaussianSigma = realParameter('maxsigma')
    desiredSNR = realParameter('desiredsnr')

    readTemplate = booleanParameter('readtemplateset')
    writeTemplate = booleanParameter('writetemplateset')
    templateSetName = stringParameter('templateset')

    withVarianceSet = booleanParameter('withvarianceset')
    varianceSetName = stringParameter('varianceset')

    withInMask = booleanParameter('withinmaskset')
    inMaskSetName = stringParameter('inmaskset')
    withOutMask = booleanParameter('withoutmaskset')
    outMaskSetName = stringParameter('outmaskset')
    writeBadMask = booleanParameter('writebadmaskset')
    badMaskSetName = stringParameter('badmaskset')
    writeCheeseMask = booleanParameter('writecheesemaskset')
    cheeseMaskSetName = stringParameter('cheesemaskset')

    withSrcList = booleanParameter('withsrclistset')
    srcListSetTabName = stringParameter('srclistset')
    selectionExpression = stringParameter('expression')
    raColName = stringParameter('racol')
    decColName = stringParameter('deccol')
    rateColName = stringParameter('ratecol')
    bkgRateColName = stringParameter('bkgratecol')
    fracBackground = realParameter('bkgfraction')

    withExpMap = booleanParameter('withexpimageset')
    expMapSetName = stringParameter('expimageset')
    expMapUseType = stringParameter('expmapuse')
    remultiplyByExpMap = booleanParameter('remultiply')

    forceCalcType = booleanParameter('forcecalctype')
    doCalcByFFT = booleanParameter('calcbyfft')
    limitOnNumDirectOps = realParameter('nopslimit')

    if (smoothType .eq. 'adaptive'&
    .and. minGaussianSigma .ge. maxGaussianSigma) then
call error('badSigmaRange', 'The value of maxsigma is less than or &
&equal to minsigma.')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine loadMask(xsize, ysize, mask)
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: xsize, ysize
    integer(int8), pointer :: mask(:,:)

    logical, pointer :: inMask(:,:), outMask(:,:)

    !-------------------------------------------------------------------
    ! Start

    if (withOutMask) then
      call loadFitsArray(outMaskSetName, 0, outMask) ! may use the
      ! untreated exposure map as the outMask.

      if (size(outMask, 1) .ne. xsize .or. size(outMask, 2) .ne. ysize)&
call error('outMaskWrongSize','')
    else
      allocate(outMask(xsize, ysize))
      outMask = .true.
    end if

    if (withInMask) then
      call loadFitsArray(inMaskSetName, 0, inMask) ! low and steep parts
      ! of the exposure map should be excluded from the inMask.

      if (size(inMask, 1) .ne. xsize .or. size(inMask, 2) .ne. ysize)&
call error('inMaskWrongSize','')
    else
      allocate(inMask(xsize, ysize))
      inMask = .true. 
    end if

    allocate(mask(xsize, ysize))
    mask = 0
    where(inMask)  mask = ibset(mask, 0)
    where(outMask) mask = ibset(mask, 1)

    deallocate(inMask, outMask)

  end subroutine loadMask

  !---------------------------------------------------------------------
  subroutine addSrcMask(mask)
  !---------------------------------------------------------------------

use dump_mod

    implicit none

    integer(int8), intent(inout) :: mask(xStart:xFinis, yStart:yFinis)

    type(DataSetT) :: imageSet, sourceSet
    type(ArrayT) :: inArray
    type(TableT) :: sourceTab
    integer :: posOfColon
    character(255) :: srcListSetName
    character(20) :: srcListTabName
    integer(int32) :: numFilteredRows, src
    type(wcsDataType) :: srcWcs
    real(single), pointer :: srcXpixels(:), srcYpixels(:),&
      srcCutoutXampPixels(:), srcCutoutYampPixels(:), srcCutoutPhase(:)
    logical, pointer ::&
      srcIsValid(:)
    real(single) :: maxRadius, x, y
    integer(int16) ::&
      xi, yi, xiMin, xiMax, yiMin, yiMax

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading source list and making &
    &sources mask.')

!........................................................................
! Get image wcs info and set up the cal:

    imageSet = dataSet(inSetName, READ)
    inArray = array(imageSet, 0)
    call getCalInfo(imageSet, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! sets CAL instrument state.

    if (instrumentId .ne. EMOS1&
    .and. instrumentId .ne. EMOS2&
    .and. instrumentId .ne. EPN) then
call error('badInstrum', 'Instrument '//trim(instId(instrumentId))//&
&' not supported.')
    end if

    call getWcsInfo(imageSet, inArray, wcs)

    call release(imageSet)

!........................................................................
! Load and filter source list:

    posOfColon = index(srcListSetTabName, ':')
    if (posOfColon .le. 0) then ! no colon in the set-tab name. User
    ! has forgotten to include the table name.
      call warning('missingTabName', "No colon+table name found after &
        &the source list set name. I'm going to use the first table in &
        &the set.")
      srcListSetName = srcListSetTabName
      srcListTabName = name(table(dataSet(srcListSetName, READ), 0))

    else
      srcListSetName = srcListSetTabName(:posOfColon - 1)
      srcListTabName = srcListSetTabName(posOfColon + 1:)
    end if

    sourceSet = clone(srcListSetName, 'tempset.ds', TEMP)
    sourceTab = table(sourceSet, srcListTabName, MODIFY)

    call message(AppMsg, Verbose, 'Selection of sources.')

    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourceTab)

    if (numFilteredRows .le. 0) then
      call warning('noSourcesSelected'&
      , 'No sources passed the specified filtering expression.')
  return
    end if

    srcWcs%X%refPixels       = int32Attribute(sourceSet, 'TCRPIX1')
    srcWcs%Y%refPixels       = int32Attribute(sourceSet, 'TCRPIX2')
    srcWcs%X%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT1')
    srcWcs%Y%pixelSizeWorld = real32Attribute(sourceSet, 'TCDELT2')
    srcWcs%X%refWorld       = real32Attribute(sourceSet, 'TCRVAL1')
    srcWcs%Y%refWorld       = real32Attribute(sourceSet, 'TCRVAL2')
    srcWcs%X%ctype          = stringAttribute(sourceSet, 'TCTYPE1')
    srcWcs%Y%ctype          = stringAttribute(sourceSet, 'TCTYPE2')

call getSourceCutouts(sourceTab, wcs, srcWcs, scAttitude&
, expStartTimeStamp, fracBackground, raColName, decColName&
, rateColName, bkgRateColName, srcXpixels, srcYpixels&
, srcCutoutXampPixels, srcCutoutYampPixels, srcCutoutPhase, srcIsValid)

    call release(sourceSet)

!**** squawk if size(srcXpixels) \= numFilteredRows.

    do src = 1, numFilteredRows
    if (.not. srcIsValid(src)&
    .or. srcCutoutXampPixels(src) .le. 0.0&
    .or. srcCutoutYampPixels(src) .le. 0.0) cycle

      maxRadius = max(srcCutoutXampPixels(src), srcCutoutYampPixels(src))
      xiMin =   floor(srcXpixels(src) - maxRadius)
      xiMax = ceiling(srcXpixels(src) + maxRadius)
      yiMin =   floor(srcYpixels(src) - maxRadius)
      yiMax = ceiling(srcYpixels(src) + maxRadius)

      xiMin = max(xiMin, xStart)
      xiMax = min(xiMax, xFinis)
      yiMin = max(yiMin, yStart)
      yiMax = min(yiMax, yFinis)

      do xi = xiMin, xiMax
        x = real(xi) - srcXpixels(src)
        do yi = yiMin, yiMax
          y = real(yi) - srcYpixels(src)
          if (pointInEllipseWithPhase(x, y, srcCutoutXampPixels(src)&
          , srcCutoutYampPixels(src), srcCutoutPhase(src)) .le. 0)&
!            inMask(xi, yi) = .false.
            mask(xi, yi) = ibclr(mask(xi, yi), 0)
        end do ! yi
      end do ! xi
    end do ! src

    deallocate(srcXpixels, srcYpixels, srcCutoutXampPixels&
      , srcCutoutYampPixels, srcCutoutPhase)

call dumpImageToFits((btest(mask, 0)), 'cheese.ds')

  end subroutine addSrcMask

  !---------------------------------------------------------------------
  subroutine divideImages(denomImage, numImage, numAberrantCounts)
  !---------------------------------------------------------------------
  ! Divides numImage by denomImage and leaves the result in numImage. If
  ! the denominator = 0, leaves 0 in the result. Calculates and optionally
  ! returns the number of pixels for which the denominator = 0 but not
  ! the numerator.

    implicit none

    real(single), intent(in) ::&
      denomImage(:,:)
    real(single), intent(inout) ::&
      numImage(:,:)
    integer(int32), intent(out), optional :: numAberrantCounts

    !-------------------------------------------------------------------
    ! Start:

    where(denomImage .ne. 0.0)
      numImage = numImage / denomImage
    elsewhere
      numImage = 0.0
    end where

    if (present(numAberrantCounts))&
      numAberrantCounts = count(denomImage .eq. 0.0 .and. numImage .ne. 0.0)

  end subroutine divideImages

  !---------------------------------------------------------------------
  subroutine calcNumDirectOps(sigma, numPixelsInMask, numDirectOps&
  , lnNumDirectOps, overflow)
  !---------------------------------------------------------------------
  ! Calculates the number of multiplications that would be needed to directly
  ! convolve the portion of the image indicated by mask with the gaussian
  ! convolver of given width sigma.

    implicit none

    real(kind=single), intent(in) ::&
      sigma
    integer(kind=int32), intent(in) ::&
      numPixelsInMask
    integer(int32), intent(out) ::&
      numDirectOps
    real(single), intent(out) ::&
      lnNumDirectOps
    logical, intent(out) ::&
      overflow

    integer(kind=int32) ::&
      numConvolverPixels
    integer(kind=int16) ::&
      fullWidthPixels

    !-------------------------------------------------------------------
    ! Start:

    fullWidthPixels = 1 + 2 * convolverArrayHalfWidth(sigma)
    numConvolverPixels = fullWidthPixels * fullWidthPixels

    lnNumDirectOps = log(real(numPixelsInMask)) + log(real(numConvolverPixels))
    if (lnNumDirectOps .gt. log(0.99 * huge(numDirectOps))) then
    ! The 0.99 converts the argument of the log() function to real as is
    ! required, but 0.99 rather than eq 1.0 also allows a bit of leeway,
    ! to make sure that no number near to huge(numDirectOps) can
    ! nonetheless slip through the gap during the conversion to real.

      overflow = .true.
      numDirectOps = -1
    else
      overflow = .false.
      numDirectOps = numPixelsInMask * numConvolverPixels
    end if
  end subroutine calcNumDirectOps

  !---------------------------------------------------------------------
  function convolverArrayHalfWidth(gaussianSigma)
  !---------------------------------------------------------------------
  ! This is a function because it is called in a couple of places, and we
  ! may want to change the factor of 5.0 some day.

    implicit none

    real(kind=single), intent(in) ::&
      gaussianSigma
    integer(kind=int16) :: convolverArrayHalfWidth

    !-------------------------------------------------------------------
    ! Start:

    convolverArrayHalfWidth = floor(5.0 * gaussianSigma)

  end function convolverArrayHalfWidth

  !---------------------------------------------------------------------
  subroutine simpleSmooth(image, mask, sigma, smoothedImage)
  !---------------------------------------------------------------------
  ! Smooth whole image with the same convolver.

    implicit none

    real(single), intent(inout) ::&
      image(:,:)
    integer(int8), intent(inout) ::&
      mask(:,:)
    real(kind=single), intent(in) ::&
      sigma
    real(kind=single), intent(out) ::&
      smoothedImage(:,:)

    integer(int16) ::&
      convolverWidth
    integer ::&
      localXsize,&
      localYsize,&
      paddedXsize,&
      paddedYsize
    complex(kind=single), dimension(:,:), allocatable ::&
      imageFFT,&
      inMaskFFT
    real(single) ::&
      weightImage(size(image, 1), size(image, 2))
    real(single), pointer ::&
      expMapImage(:,:)
!    logical :: badOutMaskFlag

    !-------------------------------------------------------------------
    ! Start:

    localXsize = size(image, 1)
    localYsize = size(image, 2)

    if (withExpMap) then
      call loadFitsArray(expMapSetName, 0, expMapImage)
      call selectImageSubset(expMapImage, xStart, xFinis, yStart, yFinis)
      call divideImages(expMapImage, image)
      deallocate(expMapImage)
    end if

    if (.not. (forceCalcType .and. .not. doCalcByFFT)) then

      call message(AppMsg, Verbose, 'Simple smoothing using FFT.')

      ! Determine size of pad needed; round this to nearest multiple
      ! of 2, 3, 5 & 7:
      convolverWidth = 2 * convolverArrayHalfWidth(sigma)
      paddedXsize = localXsize + convolverWidth
      paddedXsize = findNextHighest2357multiple(paddedXsize) ! fftw_aux_mod
      paddedYsize = localYsize + convolverWidth
      paddedYsize = findNextHighest2357multiple(paddedYsize) ! fftw_aux_mod

      call fftw2d_f77_create_plan(fftForwardPlanNum, paddedXsize&
      , paddedYsize, FFTW_FORWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

      forwardPlanExists = .true.

      call fftw2d_f77_create_plan(fftReversePlanNum, paddedXsize&
      , paddedYsize, FFTW_BACKWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

      reversePlanExists = .true.

      allocate(imageFFT(paddedXsize, paddedYsize))

      imageFFT = (0.0, 0.0) ! array op, default
      imageFFT(1:localXsize, 1:localYsize) = cmplx(image, 0.0, single)
      call fftwnd_f77_one(fftForwardPlanNum, imageFFT, 0)
      if (sigma .gt. 0.0) then
        call multiplyByGaussian(imageFFT, sigma)
      end if
      call fftwnd_f77_one(fftReversePlanNum, imageFFT, 0)
      smoothedImage = real(imageFFT(1:localXsize,1:localYsize))&
/ real(paddedXsize*paddedYsize)
!??? the divide shouldn't be necessary. Investigate.
deallocate(imageFFT)

      allocate(inMaskFFT(paddedXsize, paddedYsize))
      inMaskFFT = (0.0, 0.0) ! array op, default
!      where (inMask) inMaskFFT(1:localXsize, 1:localYsize) = (1.0, 0.0)
      where (btest(mask, 0)) inMaskFFT(1:localXsize, 1:localYsize) = (1.0, 0.0)

      call fftwnd_f77_one(fftForwardPlanNum, inMaskFFT, 0)
      if (sigma .gt. 0.0) then
        call multiplyByGaussian(inMaskFFT, sigma)
      end if
      call fftwnd_f77_one(fftReversePlanNum, inMaskFFT, 0)
      weightImage = real(inMaskFFT(1:localXsize,1:localYsize))&
 / real(paddedXsize*paddedYsize)
!??? the divide shouldn't be necessary. Investigate.
deallocate(inMaskFFT)

      call fftwnd_f77_destroy_plan(fftForwardPlanNum)
      forwardPlanExists = .false.
      call fftwnd_f77_destroy_plan(fftReversePlanNum)
      reversePlanExists = .false.

      where (btest(mask, 1))
        where (weightImage >= minAllowedWeight)
          smoothedImage = smoothedImage / weightImage
        elsewhere
!          smoothedImage = 0.0 ! not necessary - this is done globally later.
          mask = ibset(mask, 2) ! indicates those pixels where the weight
          ! is too small
        end where
      elsewhere
        smoothedImage = 0.0
      end where

    else
      call message(AppMsg, Verbose, 'Simple smoothing by direct convolution.')

      smoothedImage = 0.0
      call calcByDirect(sigma, image, mask, smoothedImage)
    end if

!if (badOutMaskFlag)&
!  call warning('outMaskTooBroad','')

    if (withExpMap .and. remultiplyByExpMap) then
!      call loadExpMap(expMapImage)
call loadFitsArray(expMapSetName, 0, expMapImage)
call selectImageSubset(expMapImage, xStart, xFinis, yStart, yFinis)
      smoothedImage = smoothedImage * expMapImage
      deallocate(expMapImage)
    end if

  end subroutine simpleSmooth

  !---------------------------------------------------------------------
  subroutine multiplyByGaussian(workPiece, gaussianSigma)
  !---------------------------------------------------------------------
  ! The reason this routine is so finicky is to optimise for speed. 

    implicit none

    real(single), intent(in) ::&
      gaussianSigma
    complex(single), intent(inout) ::&
      workPiece(:,:)

    real(single) ::&
      ffsigma_wx, ffsigma_wy
    real(single), dimension(size(workPiece, 1), size(workPiece, 2)) ::&
      convolver
    integer(int16) :: xi, yi, halfX, halfY, xTrunc, yTrunc,&
      localXsize,&
      localYsize

    !-------------------------------------------------------------------
    ! Start:

    localXsize = size(workPiece, 1)
    localYsize = size(workPiece, 2)

    convolver = 0.0

    ffsigma_wx = localXsize / TwoPi / gaussianSigma
    ffsigma_wy = localYsize / TwoPi / gaussianSigma
    xTrunc = ceiling(ffsigma_wx * 5.0)
    yTrunc = ceiling(ffsigma_wy * 5.0)
!**** have this factor of 5 as a global parameter??

    halfX = localXsize / 2 + 1
    halfY = localYsize / 2 + 1
    xTrunc = min(xTrunc, halfX)
    yTrunc = min(yTrunc, halfY)

    xi = 1
    yi = 1
    convolver(xi, yi) = calcFoldedGaussian(xi, yi, localXsize, localYsize&
    , ffsigma_wx, ffsigma_wy)
    do yi = 2, yTrunc
      convolver(xi, yi) = calcFoldedGaussian(xi, yi, localXsize, localYsize&
      , ffsigma_wx, ffsigma_wy)
      convolver(xi, localYsize + 2 - yi) = convolver(xi, yi)
    end do

    do xi = 2, xTrunc
      yi = 1
      convolver(xi, yi) = calcFoldedGaussian(xi, yi, localXsize, localYsize&
      , ffsigma_wx, ffsigma_wy)
      convolver(localXsize + 2 - xi, yi) = convolver(xi, yi)
      do yi = 2, yTrunc
        convolver(xi, yi) = calcFoldedGaussian(xi, yi, localXsize, localYsize&
        , ffsigma_wx, ffsigma_wy)
        convolver(xi, localYsize + 2 - yi) = convolver(xi, yi)
        convolver(localXsize + 2 - xi, yi) = convolver(xi, yi)
        convolver(localXsize + 2 - xi, localYsize + 2 - yi) = convolver(xi, yi)
      end do
    end do

    workPiece = workPiece * convolver ! array op.

  end subroutine multiplyByGaussian

  !---------------------------------------------------------------------
  function calcFoldedGaussian(xi, yi, localXsize, localYsize, ffsigma_wx&
  , ffsigma_wy) result(result)
  !---------------------------------------------------------------------
  ! The gaussian convolver in real space is of course also a gaussian in
  ! Fourier space. Cyclic wrapping in real space is avoided by making the
  ! images large enough but cannot be avoided in Fourier space. The true
  ! FFT of a gaussian convolver is therefore cyclically wrapped. This
  ! subroutine calculates these true, wrapped values of the FFTd convolver.
  ! Whether this is more efficient than calculating the convolver in real
  ! space then FFTing is hard to say.

    implicit none

    integer(int16), intent(in) ::&
      localXsize,&
      localYsize,&
      xi,&
      yi
    real(single), intent(in) ::&
      ffsigma_wx, ffsigma_wy
    integer(int16) ::&
      maxXfolds, maxYfolds, xFold, yFold
    real(single) :: result, x, x2, y

    !-------------------------------------------------------------------
    ! Start:

    result = 0.0
    maxXfolds = floor(6.0 * ffsigma_wx / real(localXsize))
    maxYfolds = floor(6.0 * ffsigma_wy / real(localYsize))
    do xFold = -maxXfolds, maxXfolds
      x = real(xi - 1 + xFold * localXsize) / ffsigma_wx
      x2 = x * x
      do yFold = -maxYfolds, maxYfolds
        y = real(yi - 1 + yFold * localYsize) / ffsigma_wy
        result = result + exp(-(x2 + y * y) / 2.0)
      end do
    end do

  end function calcFoldedGaussian

  !---------------------------------------------------------------------
  subroutine calcByDirect(sigma, inImage, mask, outImage, patchMask)
  !---------------------------------------------------------------------
  ! Convolve by direct (ie non-Fourier) process.

    implicit none

    real(single), intent(in) ::&
      sigma, inImage(:,:)
    integer(int8), intent(inout) ::&
      mask(:,:)
    real(single), intent(inout) ::&
      outImage(:,:)
!    logical, intent(inout) :: badOutMaskFlag
    logical, optional, intent(in) :: patchMask(:,:)

    integer ::&
      inStartX, inFinisX, inStartY, inFinisY,&
      outStartX, outFinisX, outStartY, outFinisY
    integer(int16) ::&
      localXsize,&
      localYsize,&
      halfWidthPixels,&
      xi,&
      yi,&
      xxi,&
      yyi
    real(double), dimension(:,:), pointer ::&
      convolver
    real(kind=single) ::&
      summ,&
      weight
    logical :: localPatchMask(size(mask, 1), size(mask, 2))
logical :: testFlag = .false.
!integer :: txi, tyi

    !-------------------------------------------------------------------
    ! Start:

if (present(patchMask)) then
  localPatchMask = (btest(mask, 1) .and. patchMask)
else
  localPatchMask = btest(mask, 1)
end if

!  if (.not. any(outMask)) return
!  if (.not. any(btest(mask, 1))) return
  if (.not. any(localPatchMask)) return

    localXsize = size(inImage, 1)
    localYsize = size(inImage, 2)

    halfWidthPixels = convolverArrayHalfWidth(sigma)
    if (sigma .le. 0.0 .or. halfWidthPixels .le. 0) then ! don't smooth
    ! at all.
!      where (outMask) outImage = inImage
!      where (btest(mask, 1)) outImage = inImage
      where (localPatchMask) outImage = inImage
  return
    end if

    call calculateConvolver(sigma, halfWidthPixels, convolver)
    convolver = convolver / sum(convolver)

!call maskArrayLimits(inMask, inStartX, inFinisX, inStartY, inFinisY)
!call maskArrayLimits(outMask, outStartX, outFinisX, outStartY, outFinisY)
call maskArrayLimits(btest(mask, 0), inStartX, inFinisX, inStartY, inFinisY)
call maskArrayLimits(btest(mask, 1), outStartX, outFinisX, outStartY, outFinisY)

    ! Now the actual convolution:
    do xi = outStartX, outFinisX
      do yi = outStartY, outFinisY
!      if (.not. outMask(xi, yi)) cycle
!      if (.not. btest(mask(xi, yi), 1)) cycle
      if (.not. localPatchMask(xi, yi)) cycle

        ! The pixel (xi, yi) has been established as being within the outMask
        ! and therefore it is a valid pixel for which to calculate the
        ! convolution sum.

        summ = 0.0
        weight = 0.0
        do xxi = xi - 1 - halfWidthPixels, xi + 1 + halfWidthPixels
        if (xxi .le. inStartX .or. xxi .gt. inFinisX) cycle
          do yyi = yi - 1 - halfWidthPixels, yi + 1 + halfWidthPixels
          if (yyi .lt. inStartY .or. yyi .gt. inFinisY) cycle
!          if (.not. inMask(xxi, yyi)) cycle
          if (.not. btest(mask(xxi, yyi), 0)) cycle

            ! The pixel (xxi, yyi) has been established as lying within
            ! the inMask and is therefore available to be included in the
            ! convolution sum for pixel (xi, yi).

            summ = summ + convolver(xxi - xi, yyi - yi) * inImage(xxi, yyi)
            weight = weight + convolver(xxi - xi, yyi - yi)
          end do
        end do

        if (weight >= minAllowedWeight) then
          outImage(xi, yi) = summ / weight
        else
!          badOutMaskFlag = .true. ! outImage is already 0 by default,
          mask(xi, yi) = ibset(mask(xi, yi), 2) ! flags pixels where the weight
          ! is too small.

!          outImage(xi, yi) = 0.0 ! not necessary - this is done globally later.
        end if
      end do
    end do

    deallocate(convolver)

if (testFlag) then
!txi = 331
!tyi = 350
print *, 'badmask, inMask, patchMask & outImage at (',txi,',',tyi,'): '&
, btest(mask(txi - xStart + 1, tyi - yStart + 1), 2)&
, btest(mask(txi - xStart + 1, tyi - yStart + 1), 0)&
, localPatchMask(txi - xStart + 1, tyi - yStart + 1)&
, outImage(txi - xStart + 1, tyi - yStart + 1)
end if

  end subroutine calcByDirect

  !---------------------------------------------------------------------
  subroutine calculateConvolver(sigma, halfWidthPixels, convolver)
  !---------------------------------------------------------------------
    
    implicit none

    real(single), intent(in) :: sigma
    integer(int16), intent(in) :: halfWidthPixels
    real(double), dimension(:,:), pointer ::&
      convolver

    integer(int16) ::&
      xi,&
      yi
    real(double) ::&
      sigma22,&
      x2,&
      y2

    !-------------------------------------------------------------------
    ! Start:

    sigma22 = 2.0 * sigma * sigma
    allocate(convolver(-(1 + halfWidthPixels):(1 + halfWidthPixels)&
    , -(1 + halfWidthPixels):(1 + halfWidthPixels)))
    convolver(0, 0) = 1.0
    do yi = 1, 1 + halfWidthPixels
      y2 = yi * real(yi)
      convolver(0, yi) = exp(-y2 / sigma22)
      convolver(0, -yi) = convolver(0, yi)
    end do
    do xi = 1, 1 + halfWidthPixels
      x2 = xi * real(xi)
      convolver(xi, 0) = exp(-x2 / sigma22)
      convolver(-xi, 0) = convolver(xi, 0)
      do yi = 1, 1 + halfWidthPixels
        y2 = yi * real(yi)
        convolver(xi, yi) = exp(-(x2 + y2) / sigma22)

        convolver(-xi, yi)  = convolver(xi, yi)
        convolver(-xi, -yi) = convolver(xi, yi)
        convolver(xi, -yi)  = convolver(xi, yi)
      end do
    end do

  end subroutine calculateConvolver

  !---------------------------------------------------------------------
  function convolverNorm(sigma, halfWidthPixels) result(norm)
  !---------------------------------------------------------------------
    
    implicit none

    real(single),   intent(in) :: sigma
    integer(int16), intent(in) :: halfWidthPixels
    real(single)               :: norm

    real(double), dimension(:,:), pointer :: convolver

    !-------------------------------------------------------------------
    ! Start:

    if (sigma .gt. 0.0) then
      call calculateConvolver(sigma, halfWidthPixels, convolver)
      norm = sum(convolver)
      deallocate(convolver)

    else
      norm = 1.0
    end if

  end function convolverNorm

  !---------------------------------------------------------------------
  function convolverSquareNorm(sigma, halfWidthPixels) result(squareNorm)
  !---------------------------------------------------------------------
    
    implicit none

    real(single),   intent(in) :: sigma
    integer(int16), intent(in) :: halfWidthPixels
    real(single)               :: squareNorm

    real(double), dimension(:,:), pointer :: convolver

    !-------------------------------------------------------------------
    ! Start:

    if (sigma .gt. 0.0) then
      call calculateConvolver(sigma, halfWidthPixels, convolver)
      squareNorm = sum(convolver * convolver)
      deallocate(convolver)

    else
      squareNorm = 1.0
    end if

  end function convolverSquareNorm

  !---------------------------------------------------------------------
  subroutine adaptiveSmooth(inImage, mask, smoothedImage)
  !---------------------------------------------------------------------
  ! Adaptively-smooth the image.

    implicit none

    real(single), intent(inout) ::&
      inImage(:,:)
    integer(int8), intent(inout) ::&
      mask(:,:)
    real(single), intent(out) ::&
      smoothedImage(:,:)

    integer(int8), dimension(:,:), allocatable ::&
      indexImage
    complex(single), dimension(:,:), allocatable ::&
      inImageFFT,&
      inMaskFFT,&
      varianceImageFFT
    real(single), dimension(:,:), allocatable ::&
      bestSnr, secondBestSnr, rmsSmoothedImage
    real(single), dimension(:,:), pointer ::&
      varianceImage, expMapImage
    type(gaussData), dimension(numGaussians) ::&
      gaussian
    integer(int8) ::&
      i
    integer(int16) ::&
      localXsize,&
      localYsize,&
      paddedXsize,&
      paddedYsize
    integer(int32) ::&
      numPixels,&
      numMaskedPixels
      ! with a 78x78 convolver will do so.
    real(single) :: sigma, norm, normCorrector
    integer :: status
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    numMaskedPixels = count(btest(mask, 1))
    localXsize = size(inImage, 1)
    localYsize = size(inImage, 2)

!........................................................................
! Set up the variance image:

    if (withVarianceSet) then
      call loadFitsArray(varianceSetName, 0, varianceImage)
      call selectImageSubset(varianceImage, xStart, xFinis, yStart, yFinis)

    else
      if (any(inImage .lt. 0.0))&
call error('somePixelsNegative', 'The image contains some values less than &
&zero. Because of this it is not possible to calculate a variance image. &
&You should supply one via the appropriate task parameters.')

!*** other tests to see if inImage is poissonian?
      allocate(varianceImage(localXsize, localYsize))
      varianceImage = inImage
    end if

!........................................................................
! Divide by expmap if that is desired:

    if (withExpMap) then
      call loadFitsArray(expMapSetName, 0, expMapImage)
      call selectImageSubset(expMapImage, xStart, xFinis, yStart, yFinis)
      call divideImages(expMapImage, inImage)
      if (expMapUseType .eq. 'samesnr') then
        call divideImages(expMapImage * expMapImage, varianceImage)
      else if (expMapUseType .eq. 'samesize') then
        call divideImages(expMapImage * maxval(expMapImage), varianceImage)
      else
call error('badExpMapUse', 'The value '//trim(expMapUseType)//' of &
&--expmapuse is not recognised.')
      end if
      deallocate(expMapImage)
    end if

!........................................................................
! Calculate some FFTs and set up the library of gaussians:

    call calcPaddedSizes(maxGaussianSigma, paddedXsize, paddedYsize)
    allocate(inImageFFT(paddedXsize, paddedYsize)&
    , inMaskFFT(paddedXsize, paddedYsize)&
    , varianceImageFFT(paddedXsize, paddedYsize))

    if (.not. (forceCalcType .and. .not. doCalcByFFT)) then
      call prepareFFT(inImage, btest(mask, 0), varianceImage, inImageFFT&
      , inMaskFFT, varianceImageFFT)
    end if
    call calcGaussianLibrary(gaussian)
    maxSnrImprovement = gaussian(1)%snrImprovement

    allocate(indexImage(localXsize, localYsize))

!........................................................................
! Calculate the index image:

    indexImage = 0
    if (readTemplate) then
      call loadTemplateAndCreateIndexImage(gaussian, indexImage)
      where(.not. btest(mask, 1))
        indexImage = 0
      end where

    else
      call message(AppMsg, Verbose, 'Preparing template:')

      allocate(bestSnr(localXsize, localYsize)&
      , secondBestSnr(localXsize, localYsize)&
      , rmsSmoothedImage(localXsize, localYsize))

      where(btest(mask, 1))
        indexImage = 1
      elsewhere
        indexImage = 0
      end where

!call dumpImageToFits(indexImage, 'junk.ds')

      smoothedImage = 0.0
      rmsSmoothedImage = 0.0
      i = 1

      sigma = gaussian(i)%sigma
      numPixels = count(indexImage .eq. i)
      gaussian%numPixels = numPixels

      status = maskedSmooth(i, sigma, numPixels, numMaskedPixels, 1&
        , mask, inImage, inImageFFT, inMaskFFT, smoothedImage)

      if (status == 1)&
call error('wholeImageUnmasked', "The number of mask pixels is 0. &
&Can't smooth masked image.")

      status = maskedSmooth(i, real(sigma / RootTwo, single), numPixels&
        , numMaskedPixels, 0, mask, varianceImage, varianceImageFFT, inMaskFFT&
        , rmsSmoothedImage)

      norm = convolverNorm(sigma, convolverArrayHalfWidth(sigma))
      normCorrector = convolverSquareNorm(sigma&
        , convolverArrayHalfWidth(sigma)) / norm / norm

      where(indexImage .eq. i)&
        rmsSmoothedImage = sqrt(rmsSmoothedImage * normCorrector)

      where(rmsSmoothedImage .gt. 0.0 .and. smoothedImage .gt. 0.0)
        bestSnr = smoothedImage / rmsSmoothedImage
      elsewhere
        bestSnr = 0.0
      end where
!call dumpImageToFits(indexImage, 'g1_index.ds')
!call dumpImageToFits(smoothedImage, 'g1_smooth.ds')
!call dumpImageToFits(rmsSmoothedImage, 'g1_rms.ds')
!call dumpImageToFits(bestSnr, 'g1_best.ds')

!where(btest(mask, 2)) mask = ibset(mask, 3) ! marks the incorrigibles that !occurred at the maximum level of smoothing.

if (testFlag) print *, 'num incorrigibles:', count(btest(mask, 2))

      where(indexImage .eq. i .and. bestSnr .gt. desiredSnr&
      .and. .not. btest(mask, 2))&
!      where(indexImage .eq. i .and. bestSnr .gt. desiredSnr)&
        indexImage = 2 ! 2 here is ok because there must be at least 2
      ! gaussians for adaptive smoothing. 'Adaptive smoothing' with
      ! only 1 gaussian is just the same thing as simple smoothing.

      do i = 2, size(gaussian)
        secondBestSnr = bestSnr

        sigma = gaussian(i)%sigma
        numPixels = count(indexImage .eq. i)
        gaussian%numPixels = numPixels

        status = maskedSmooth(i, sigma, numPixels, numMaskedPixels, 1, mask&
          , inImage, inImageFFT, inMaskFFT, smoothedImage, (indexImage .eq. i))

        if (status == 1) then
          write(errstr, '(a, f6.2, a)') 'No gaussian with sigma less than '&
            , sigma, ' is required..'
          call message(AppMsg, Noisy)
      exit
        end if

        status = maskedSmooth(i, real(sigma / RootTwo, single), numPixels&
          , numMaskedPixels, 0, mask, varianceImage, varianceImageFFT&
          , inMaskFFT, rmsSmoothedImage, (indexImage .eq. i))

        norm = convolverNorm(sigma, convolverArrayHalfWidth(sigma))
        normCorrector = convolverSquareNorm(sigma&
        , convolverArrayHalfWidth(sigma)) / norm / norm

if (testFlag) print *, '  indexImage(',txi,',',tyi,'): ', indexImage(txi - xStart + 1, tyi - yStart + 1)

        where(indexImage .eq. i)
          rmsSmoothedImage = sqrt(rmsSmoothedImage * normCorrector)
          where(rmsSmoothedImage .gt. 0.0 .and. smoothedImage .gt. 0.0)
            bestSnr = smoothedImage / rmsSmoothedImage
          elsewhere
            bestSnr = 0.0
          end where

          where(btest(mask, 2))
            indexImage = i - 1 ! ie, smooth this pixel harder - this ought
            ! to avoid the problem with too-low weight here.
          elsewhere
            where(bestSnr .lt. desiredSnr)
              where(abs(secondBestSnr - desiredSnr)&
              .lt. abs(bestSnr - desiredSnr))
                indexImage = i - 1 ! ie, the value that gave the best SNR.
              end where
            elsewhere
              indexImage = i + 1 ! but see just beyond the end of the loop:
            end where
          end where
        end where

if (testFlag) then
  print *, '  smoothedImage(',txi,',',tyi,'): ', smoothedImage(txi - xStart + 1, tyi - yStart + 1)
  print *, '  rmsSmoothedImage(',txi,',',tyi,'): ', rmsSmoothedImage(txi - xStart + 1, tyi - yStart + 1)
  print *, '  bestSnr(',txi,',',tyi,'): ', bestSnr(txi - xStart + 1, tyi - yStart + 1)
  print *, '  badMask(',txi,',',tyi,'): ', btest(mask(txi - xStart + 1, tyi - yStart + 1), 2)
end if

!if (i .eq. 2) then
!call dumpImageToFits(indexImage,       'g2_index.ds')
!call dumpImageToFits(smoothedImage,    'g2_smooth.ds')
!call dumpImageToFits(rmsSmoothedImage, 'g2_rms.ds')
!call dumpImageToFits(bestSnr,          'g2_best.ds')
!!call error('','')
!end if
!if (i .eq. 3) then
!call dumpImageToFits(indexImage,       'g3_index.ds')
!call dumpImageToFits(smoothedImage,    'g3_smooth.ds')
!call dumpImageToFits(rmsSmoothedImage, 'g3_rms.ds')
!call dumpImageToFits(bestSnr,          'g3_best.ds')
!call error('','')
!end if
      end do ! gaussian number
      where(indexImage .gt. size(gaussian)) indexImage = size(gaussian)
if (testFlag) then
  print *, ''
  print *, 'size(gaussian): ', size(gaussian)
  print *, '  indexImage(',txi,',',tyi,'): ', indexImage(txi - xStart + 1, tyi - yStart + 1)
  print *, '  inMask(',txi,',',tyi,'): ', btest(mask(txi - xStart + 1, tyi - yStart + 1), 0)
end if
      where(indexImage == size(gaussian) .and. .not. btest(mask, 0))&
        indexImage = size(gaussian) - 1 ! Can't let these pixels be 'smoothed'
        ! with zero-width gaussians.

if (testFlag) then
  print *, ''
  print *, '  indexImage(',txi,',',tyi,'): ', indexImage(txi - xStart + 1, tyi - yStart + 1)
  print *, '  inMask(',txi,',',tyi,'): ', btest(mask(txi - xStart + 1, tyi - yStart + 1), 0)
end if

      deallocate(rmsSmoothedImage, secondBestSnr, bestSnr, varianceImage&
      , varianceImageFFT)

      mask = ibclr(mask, 2) ! if this bit is set again for any pixel, it
      ! should be because the pixel is a real incorrigible that give too
      ! low weight even at the broadest level of smoothing.
    end if ! readTemplate

!........................................................................
! Now smooth the input according to the index image:

    call message(AppMsg, Verbose, 'Smoothing image:')

    smoothedImage = 0.0

    do i = 1, size(gaussian)
      sigma = gaussian(i)%sigma
      numPixels = count(indexImage .eq. i)
      gaussian%numPixels = numPixels

      status = maskedSmooth(i, sigma, numPixels, numMaskedPixels, 2, mask&
        , inImage, inImageFFT, inMaskFFT, smoothedImage, (indexImage .eq. i))

      if (status == 1) then
        write(errstr, '(a, i3.3, a, f6.2, a, f7.3, a)') '  ', i&
          , 'th gaussian, sigma = ', sigma, '; percent pixels = ', numPixels&
          / real(numMaskedPixels) / 0.01, ' %.'
        call message(AppMsg, Noisy)
    cycle
      end if
    end do ! gaussian number

!........................................................................
! Clean up:

    if (forwardPlanExists) then
      call fftwnd_f77_destroy_plan(fftForwardPlanNum)
      forwardPlanExists = .false.
    end if
    if (reversePlanExists) then
      call fftwnd_f77_destroy_plan(fftReversePlanNum)
      reversePlanExists = .false.
    end if

    if (writeTemplate) call calcAndSaveTemplate(gaussian, indexImage)
    deallocate(indexImage)

    if (withExpMap .and. remultiplyByExpMap) then
      call loadFitsArray(expMapSetName, 0, expMapImage)
      call selectImageSubset(expMapImage, xStart, xFinis, yStart, yFinis)
      smoothedImage = smoothedImage * expMapImage
      deallocate(expMapImage)
    end if

  end subroutine adaptiveSmooth

  !---------------------------------------------------------------------
  subroutine loadTemplateAndCreateIndexImage(gaussian, indexImage)
  !---------------------------------------------------------------------
!*** use fracVal.
   
    implicit none

    type(gaussData), intent(in) ::&
      gaussian(:)
    integer(int8), intent(inout) :: indexImage(:,:)

    integer(int8) :: i, numGauss
    real(single), pointer ::&
      templateImage(:,:)
    type(DataSetT) :: inSet
    type(ArrayT) :: inArray
    real(single) :: loSigma, hiSigma, frac, sigma, snrImpLo, snrImpHi
    character(20) :: templateStyle
    type(gaussData) :: tempGaussian(size(gaussian))
    integer(int16) :: halfWidthPixels

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading template image '&
    //trim(templateSetName)//'.')

    numGauss = size(gaussian)

call loadFitsArray(templateSetName, 0, templateImage)
call selectImageSubset(templateImage, xStart, xFinis, yStart, yFinis)

    inSet = dataSet(templateSetName, READ)
    inArray = array(inSet, 0, READ)
    if (hasAttribute(inArray, "FORMAT")) then
      templateStyle = stringAttribute(inArray, "FORMAT")
    else
      templateStyle = 'none'
    end if

!    call loadImage(inArray, templateImage, withBounds=.true.)
!    call release(inSet)

    if (templateStyle .eq. '2.7') then
      loSigma = (gaussian(1)%sigma + gaussian(2)%sigma) / 2.0
      where(templateImage .ge. loSigma) indexImage = 1
      do i = 2, numGauss - 1
        hiSigma = loSigma
        loSigma = (gaussian(i)%sigma + gaussian(i+1)%sigma) / 2.0
        where(templateImage .lt. hiSigma&
        .and.  templateImage .ge. loSigma)
          indexImage = i
        end where
      end do
      hiSigma = (gaussian(numGauss-1)%sigma + gaussian(numGauss)%sigma) / 2.0
      where(templateImage .lt. hiSigma) indexImage = numGauss

    else ! presumably templateStyle = 'none'
      call warning('oldStyleTemplate', 'An old-style template file &
      &has been detected. No cause for concern, this will be read &
      &correctly and converted to the new style upon output.')

      do i = 1, numGauss - 1
        frac = (i - 1) / real(numGauss - 2)
        sigma = minGaussianSigma * frac + maxGaussianSigma * (1.0 - frac)
!sigma = maxGaussianSigma * (1.0 - frac)
        halfWidthPixels = convolverArrayHalfWidth(sigma)
        tempGaussian(i)%snrImprovement = convolverNorm(sigma, halfWidthPixels)&
        / sqrt(convolverSquareNorm(sigma, halfWidthPixels))

!print *, 'i, sigma, real sigma, snrImp: '&
!, i, sigma, gaussian(i  )%sigma, tempGaussian(i)%snrImprovement

! Notes: (i) The improvement in SNR given by the above formula assumes that
! the average value of the function or image to be smoothed is constant. It
! is the departure of real images from this ideal that requires the iteration
! implemented in this software. (ii) For large sigma the SNR improvement
! wrought by a sampled gaussian is proportional to 2 * sigma * RootPi. As
! sigma -> 0 the SNR improvement departs from this rule and asymtotes to 1.

      end do
      ! NOTE! that tempGaussian(numGauss)%snrImprovement is not defined.


      snrImpLo = tempGaussian(1)%snrImprovement * 0.9&
               + tempGaussian(2)%snrImprovement * 0.1
      where(templateImage .ge. snrImpLo) indexImage = 1

!print *, 'i, snrImpLo, numPix: '&
!, 1, snrImpLo, count(templateImage .ge. snrImpLo)

      do i = 2, numGauss - 2
        snrImpHi = snrImpLo
        snrImpLo = tempGaussian(i)%snrImprovement
        where(templateImage .lt. snrImpHi&
        .and. templateImage .ge. snrImpLo)&
          indexImage = i
!print *, 'i, snrImpLo, snrImpHi, numPix: '&
!, i, snrImpLo, snrImpHi, count(templateImage .lt. snrImpHi&
!.and. templateImage .ge. snrImpLo)
      end do

      snrImpHi = snrImpLo
      snrImpLo = tempGaussian(numGauss - 1)%snrImprovement * 0.9&
               + tempGaussian(numGauss - 2)%snrImprovement * 0.1
      where(templateImage .lt. snrImpHi&
      .and. templateImage .ge. snrImpLo)&
        indexImage = numGauss - 1
!print *, 'i, snrImpLo, snrImpHi, numPix: '&
!, numGauss - 1, snrImpLo, snrImpHi, count(templateImage .lt. snrImpHi&
!.and. templateImage .ge. snrImpLo)

      snrImpHi = snrImpLo
      where(templateImage .lt. snrImpHi) indexImage = numGauss
!print *, 'i, snrImpHi, numPix: '&
!, numGauss, snrImpHi, count(templateImage .lt. snrImpHi)
    end if

    deallocate(templateImage)

!call dumpImageToFits(indexImage, 'junk_index')

  end subroutine loadTemplateAndCreateIndexImage

  !---------------------------------------------------------------------
  subroutine loadTemplateImage(templateImage)
  !---------------------------------------------------------------------
    
    implicit none

    real(single), intent(out) ::&
      templateImage(:,:)

    type(DataSetT) :: inSet
    type(ArrayT) :: inArray
    integer(kind=int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(kind=int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(kind=int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(kind=single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(kind=double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Reading template image '&
    //trim(templateSetName)//'.')

    inSet = dataSet(templateSetName, READ)
    inArray = array(inSet, 0, READ)

    if (dataType(inArray) .eq. INTEGER8) then
      arrayDataInt8 => int8Array2Data(inArray)
      templateImage = real(arrayDataInt8, single)
    else if (dataType(inArray) .eq. INTEGER16) then
      arrayDataInt16 => int16Array2Data(inArray)
      templateImage = real(arrayDataInt16, single)
    else if (dataType(inArray) .eq. INTEGER32) then
      arrayDataInt32 => int32Array2Data(inArray)
      templateImage = real(arrayDataInt32, single)
    else if (dataType(inArray) .eq. REAL32) then
      arrayDataReal32 => real32Array2Data(inArray)
      templateImage = arrayDataReal32
    else ! this gives a DAL error if data type is none of the above.
      arrayDataReal64 => real64Array2Data(inArray)
      templateImage = real(arrayDataReal64, single)
    end if

    call release(inSet)

  end subroutine loadTemplateImage

  !---------------------------------------------------------------------
  subroutine calcGaussianLibrary(gaussian)
  !---------------------------------------------------------------------
  ! The structure 'gaussian' is used to store five quantities for each of the
  ! gaussian convolvers in the library: the width or sigma value of the
  ! gaussian, the normalisation parameter (this is the sum of the elements
  ! of the unnormalised gaussian convolver array), the improvement in SNR
  ! associated with this gaussian, the half-width of the truncation boundary
  ! in pixels and the number of pixels assigned to be smoothed by this
  ! gaussian. Only the first three are calculated here: the fourth value
  ! can vary because it depends on the number of iterations applied to the
  ! image.
  !
  ! The gaussians in the library are ordered so that the lower the value
  ! of sigmaIndex, the broader the gaussian.
  !
  ! Image pixels for which the desired SNR improvement is greater than a
  ! given maximum are smoothed by the broadest gaussian in the library;
  ! those for which the desired improvement in SNR is less than or equal
  ! to 1 are smoothed by a zero-width gaussian (ie, not smoothed at all).

    implicit none

    type(gaussData), intent(out) :: gaussian(:)

    integer(int16) ::&
      i, gsize, halfWidthPixels
    real(single) ::&
      frac, sigma, norm

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating widths of the library &
    &of gaussians.')

    gsize = size(gaussian)
    gaussian%numPixels = 0 ! vector op, default

!      do i = 1, gsize
!        frac = (i - 1) / real(gsize - 1)
!        gaussian(i)%sigma = minGaussianSigma * frac + maxGaussianSigma&
!        * (1.0 - frac)
!        gaussian(i)%halfWidthPixels = convolverArrayHalfWidth(gaussian(i)%sigma)
!        gaussian(i)%norm = convolverNorm(gaussian(i)%sigma&
!        , gaussian(i)%halfWidthPixels)
!      end do
!      do i = 1, gsize - 1
!        frac = (i - 1) / real(gsize - 2)
!        sigma = minGaussianSigma * frac + maxGaussianSigma * (1.0 - frac)
!        halfWidthPixels = convolverArrayHalfWidth(sigma)
!        gaussian(i)%snrImprovement = convolverNorm(sigma, halfWidthPixels)&
!        / sqrt(convolverSquareNorm(sigma, halfWidthPixels))


    do i = 1, gsize
      frac = (i - 1) / real(gsize - 1)
      sigma = minGaussianSigma * frac + maxGaussianSigma * (1.0 - frac)
!sigma = maxGaussianSigma * (1.0 - frac)
      halfWidthPixels = convolverArrayHalfWidth(sigma)
      norm = convolverNorm(sigma, halfWidthPixels)

      gaussian(i)%sigma           = sigma
      gaussian(i)%halfWidthPixels = halfWidthPixels
      gaussian(i)%norm            = norm
      gaussian(i)%snrImprovement  = norm&
      / sqrt(convolverSquareNorm(sigma, halfWidthPixels))

! Notes: (i) The improvement in SNR given by the above formula assumes that
! the average value of the function or image to be smoothed is constant. It
! is the departure of real images from this ideal that requires the iteration
! implemented in this software. (ii) For large sigma the SNR improvement
! wrought by a sampled gaussian is proportional to 2 * sigma * RootPi. As
! sigma -> 0 the SNR improvement departs from this rule and asymtotes to 1.
    end do

  end subroutine calcGaussianLibrary

  !---------------------------------------------------------------------
  subroutine calcPaddedSizes(sigma, paddedXsize, paddedYsize)
  !---------------------------------------------------------------------
 
    implicit none

    real(kind=single), intent(in) ::&
      sigma
    integer(int16), intent(out) ::&
      paddedXsize,&
      paddedYsize

    integer(int16) ::&
      convolverWidth

    !-------------------------------------------------------------------
    ! Start:

    ! Determine size of pad needed; round this to nearest multiple
    ! of 2, 3, 5 & 7:
    convolverWidth = 2 * convolverArrayHalfWidth(sigma)
    paddedXsize = (1 + xFinis - xStart) + convolverWidth
    paddedXsize = findNextHighest2357multiple(int(paddedXsize)) ! fftw_aux_mod
    paddedYsize = (1 + yFinis - yStart) + convolverWidth
    paddedYsize = findNextHighest2357multiple(int(paddedYsize)) ! fftw_aux_mod

  end subroutine calcPaddedSizes

  !---------------------------------------------------------------------
  subroutine prepareFFT(inImage, inMask, varianceImage, inImageFFT, inMaskFFT&
  , varianceImageFFT)
  !---------------------------------------------------------------------
 
    implicit none

    real(kind=single), intent(in) ::&
      inImage(:,:), varianceImage(:,:)
!    integer(int8), intent(in) ::&
logical, intent(in) ::&
      inMask(:,:)
    complex(single), intent(out) ::&
      inImageFFT(:,:),&
      inMaskFFT(:,:),&
      varianceImageFFT(:,:)

    integer(int16) ::&
      localXsize,&
      localYsize,&
      paddedXsize,&
      paddedYsize

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Preparing FFTs for adaptive routine.')

    localXsize = size(inImage, 1)
    localYsize = size(inImage, 2)
    paddedXsize = size(inImageFFT, 1)
    paddedYsize = size(inImageFFT, 2)

    call fftw2d_f77_create_plan(fftForwardPlanNum, int(paddedXsize)&
    , int(paddedYsize), FFTW_FORWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

    forwardPlanExists = .true.

    inImageFFT = (0.0, 0.0) ! array op, default
    inImageFFT(1:localXsize, 1:localYsize) = cmplx(inImage, 0.0, single)
    call fftwnd_f77_one(fftForwardPlanNum, inImageFFT, 0)

    inMaskFFT = (0.0, 0.0) ! array op, default
!    where (mask .eq. 1) inMaskFFT(1:localXsize, 1:localYsize) = (1.0, 0.0)
where (inMask) inMaskFFT(1:localXsize, 1:localYsize) = (1.0, 0.0)
    call fftwnd_f77_one(fftForwardPlanNum, inMaskFFT, 0)

    varianceImageFFT = (0.0, 0.0) ! array op, default
    varianceImageFFT(1:localXsize, 1:localYsize)&
    = cmplx(varianceImage, 0.0, single)
    call fftwnd_f77_one(fftForwardPlanNum, varianceImageFFT, 0)

    call fftwnd_f77_destroy_plan(fftForwardPlanNum)
    forwardPlanExists = .false.

    call fftw2d_f77_create_plan(fftReversePlanNum, int(paddedXsize)&
    , int(paddedYsize), FFTW_BACKWARD, FFTW_ESTIMATE + FFTW_IN_PLACE)

    reversePlanExists = .true.

  end subroutine prepareFFT

  !---------------------------------------------------------------------
  integer function maskedSmooth(i, sigma, numPixels, numMaskedPixels&
  , messageType, mask, inImage, inImageFFT, inMaskFFT, smoothedImage&
  , patchMask) result(status)
  !---------------------------------------------------------------------

    implicit none

    integer(int8),     intent(in)    :: i
    integer,           intent(in)    :: messageType
    real(single),      intent(in)    :: sigma
    integer(int32),    intent(in)    :: numPixels, numMaskedPixels
    integer(int8),     intent(inout) :: mask(:,:)
    real(single),      intent(in)    :: inImage(:,:)
    complex(single),   intent(in)    :: inImageFFT(:,:), inMaskFFT(:,:)
    real(single),      intent(out)   :: smoothedImage(:,:)
    logical, optional, intent(in)    :: patchMask(:,:)

    real(single) :: lnNumDirectOps
    integer(int32) ::&
      numDirectOps ! Take care that this is not overflowed. A 600x600 array
      ! with a 78x78 convolver will do so.
    logical :: overflow

    !-------------------------------------------------------------------
    ! Start

    status = 0

    call calcNumDirectOps(sigma, numPixels, numDirectOps&
    , lnNumDirectOps, overflow)

    if (numDirectOps .le. 0 .and. .not. overflow) then
      status = 1
  return
    end if

    if (messageType == 1) then
      write(errstr, '(a, i3.3, a, f6.2)') '  ', i&
      , 'th gaussian, sigma = ', sigma
      call message(AppMsg, Verbose)
    end if

    if ((forceCalcType .and. doCalcByFFT)&
    .or. (lnNumDirectOps .gt. log(limitOnNumDirectOps)&
    .and. .not. (forceCalcType .and. .not. doCalcByFFT))) then

      if (messageType == 2) then
        write(errstr, '(a, i3.3, a, f6.2, a, f7.3, a)') '  ', i&
        , 'th gaussian, sigma = ', sigma, '; percent pixels = ', numPixels&
        / real(numMaskedPixels) / 0.01, ' %: FFT.'
        call message(AppMsg, Verbose)
      end if

      if (present(patchMask)) then
        call calcByFFT(sigma, inImageFFT, inMaskFFT, mask, smoothedImage&
          , patchMask)

      else
        call calcByFFT(sigma, inImageFFT, inMaskFFT, mask, smoothedImage)
      end if

    else
      if (messageType == 2) then
        write(errstr, '(a, i3.3, a, f6.2, a, f7.3, a)') '  ', i&
        , 'th gaussian, sigma = ', sigma, '; percent pixels = ', numPixels&
        / real(numMaskedPixels) / 0.01, ' %: Direct.'
        call message(AppMsg, Verbose)
      end if

      if (present(patchMask)) then
        call calcByDirect(sigma, inImage, mask, smoothedImage, patchMask)
      else
        call calcByDirect(sigma, inImage, mask, smoothedImage)
      end if
    end if

  end function maskedSmooth

  !---------------------------------------------------------------------
  subroutine calcByFFT(sigma, inImageFFT, inMaskFFT&
  , mask, outImage, patchMask)
  !---------------------------------------------------------------------

    implicit none

    real(single),      intent(in)    :: sigma
    complex(single),   intent(in)    :: inImageFFT(:,:),&
                                        inMaskFFT(:,:)
    integer(int8),     intent(inout) :: mask(:,:)
    real(single),      intent(inout) :: outImage(:,:)
    logical, optional, intent(in)    :: patchMask(:,:)

    real(single), dimension(size(outImage, 1), size(outImage, 2)) ::&
      tempInImage,&
      weightImage
    complex(single), dimension(size(inImageFFT, 1), size(inImageFFT, 2)) ::&
      tempInImageFFT,&
      tempInMaskFFT
    integer(int16) :: localXsize, localYsize
    logical :: localPatchMask(size(mask, 1), size(mask, 2))
logical :: testFlag = .false.
!integer :: txi, tyi
    !-------------------------------------------------------------------
    ! Start:

if (present(patchMask)) then
  localPatchMask = (btest(mask, 1) .and. patchMask)
else
  localPatchMask = btest(mask, 1)
end if

    localXsize = size(outImage, 1)
    localYsize = size(outImage, 2)

    tempInImageFFT = inImageFFT
    if (sigma .gt. 0.0) then
      call multiplyByGaussian(tempInImageFFT, sigma)
    end if

    call fftwnd_f77_one(fftReversePlanNum, tempInImageFFT, 0)
    tempInImage = real(tempInImageFFT(1:localXsize,1:localYsize), single)

    tempInMaskFFT = inMaskFFT
    if (sigma .gt. 0.0) then
      call multiplyByGaussian(tempInMaskFFT, sigma)
    end if
    call fftwnd_f77_one(fftReversePlanNum, tempInMaskFFT, 0)
    weightImage = real(tempInMaskFFT(1:localXsize,1:localYsize), single)

!if (any(weightImage < minAllowedWeight .and. outMask)) then
!  badOutMaskFlag = .true.
!end if

!    where (outMask)
    where (localPatchMask)
      where(weightImage >= minAllowedWeight)
        outImage = tempInImage / weightImage
      elsewhere
!        outImage = 0.0 ! not necessary - this is done globally later.
        mask = ibset(mask, 2) ! indicates those pixels where the weight
        ! is too small
      end where
    end where

if (testFlag) then
print *, 'badmask, inMask, patchMask & outImage at (',txi,',',tyi,'): '&
, btest(mask(txi - xStart + 1, tyi - yStart + 1), 2)&
, btest(mask(txi - xStart + 1, tyi - yStart + 1), 0)&
, localPatchMask(txi - xStart + 1, tyi - yStart + 1)&
, outImage(txi - xStart + 1, tyi - yStart + 1)
end if

  end subroutine calcByFFT

  !---------------------------------------------------------------------
  subroutine calcAndSaveTemplate(gaussian, indexImage)
  !---------------------------------------------------------------------
    
    implicit none

    type(gaussData), intent(in) ::&
      gaussian(:)
    integer(int8), intent(in) :: indexImage(:,:)

    integer(int8) :: i
    real(single), allocatable ::&
      templateImage(:,:)
    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing template image '&
    //trim(templateSetName)//'.')

    allocate(templateImage(size(indexImage, 1), size(indexImage, 2))&
    , dimVector(0:1))

    templateImage = 1.0 ! the minimum sensible value. This will remain the
    ! value for all those pixels for which indexImage = 0, which amounts to
    ! those for which mask = 0.

    do i = 1, size(gaussian)
      where(indexImage .eq. i) templateImage = gaussian(i)%sigma
    end do

!****** use saveOutImage???

    outSet = dataSet(templateSetName, CREATE)
!print *, 'xsize, ysize, xStart, xFinis, yStart, yFinis: '&
!, xsize, ysize, xStart, xFinis, yStart, yFinis
!print *, 'lbound(templateImage, 1), ubound(templateImage, 1), lbound(templateImage, 2), ubound(templateImage, 2): '&
!, lbound(templateImage, 1), ubound(templateImage, 1), lbound(templateImage, 2), ubound(templateImage, 2)
    dimVector(0) = xsize
    dimVector(1) = ysize
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, '', '', 0)
    call setAttribute(outArray, "FORMAT", "2.7", "Refers to the way in &
    &which the template values are defined.")

    arrayDataReal32 => real32Array2Data(outArray)
!print *, 'lbound(arrayDataReal32, 1), ubound(arrayDataReal32, 1), lbound(arrayDataReal32, 2), ubound(arrayDataReal32, 2): '&
!, lbound(arrayDataReal32, 1), ubound(arrayDataReal32, 1), lbound(arrayDataReal32, 2), ubound(arrayDataReal32, 2)
    arrayDataReal32 = 1.0 ! the minimum sensible value.
    arrayDataReal32(xStart-1:xFinis-1, yStart-1:yFinis-1) = templateImage
    call release(outSet)

    deallocate(dimVector, templateImage)

  end subroutine calcAndSaveTemplate

  !---------------------------------------------------------------------
  subroutine saveMask(mask, newImageName&
  , xsize, ysize, xStart, yStart)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) ::&
      mask(:,:)
    character(*), intent(in) ::&
      newImageName
    integer(int16), intent(in), optional :: xsize, ysize, xStart, yStart

    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) ::&
      xsizeLocal, ysizeLocal, xStartLocal, yStartLocal,&
      xFinis, yFinis
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    if (present(xsize)) then
      xsizeLocal = xsize
    else
      xsizeLocal = size(mask, 1)
    end if
    if (present(ysize)) then
      ysizeLocal = ysize
    else
      ysizeLocal = size(mask, 2)
    end if
    if (present(xStart)) then
      xStartLocal = xStart
    else
      xStartLocal = 1
    end if
    if (present(yStart)) then
      yStartLocal = yStart
    else
      yStartLocal = 1
    end if

if (testFlag) then
print *, 'xsizeLocal, ysizeLocal, xStartLocal, yStartLocal: '&
, xsizeLocal, ysizeLocal, xStartLocal, yStartLocal
if (testFlag) print *, 'xsize: ', xsize
if (testFlag) print *, 'ysize: ', ysize
if (testFlag) print *, 'xStart: ', xStart
if (testFlag) print *, 'yStart: ', yStart
end if

!    inSet = dataSet(imageName, READ)
!    inArray = array(inSet, 0, READ)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = xsizeLocal
    dimVector(1) = ysizeLocal
    xFinis = xStartLocal + size(mask, 1) - 1
    yFinis = yStartLocal + size(mask, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', INTEGER8, dimVector, position=0)
!    call setAttributes(attributable(outSet), attributable(inSet))

!    call copyVitalArrayKwds(inArray, outArray)
!    ! This is done instead of a total copy of all array-header keywords because
!    ! doing the latter stuffs up the DSS.

    arrayDataInt8 => int8Array2Data(outArray)
    arrayDataInt8 = 0.0
    where(mask)&
      arrayDataInt8(xStartLocal-1:xFinis-1, yStartLocal-1:yFinis-1) = 1.0

    call release(outSet)
!    call release(inSet)

    deallocate(dimVector)

  end subroutine saveMask

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module asmooth_mod
