!+
! Module: coordinate
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains several routines to do with coordinate transformations.
! 
! Comments: TODO:
! 
! $Id: coordinate.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module coordinate

  use dal
  use cal	! contains 'use caloaldefs' and 'use caltypes'
  use oal	! contains 'use oaltypes'. For OAL_toAttitudeMatrix.
  use errorHandling

  use constants
  use mathutils
  use cal_aux 	! for getInstrumentAttitude.

  implicit none 

  private 

  real(single), parameter :: detUnitArcsec = 0.05 ! this should be centrally accessible somewhere? (But where?? In a parameter file?)

  public raDecToDetXY
  public detXYtoRaDec
  public raDecToTan
  public tanToRaDec
  public detXYtoTan

  interface detXYtoRaDec
    module procedure detXYtoRaDecScalar
    module procedure detXYtoRaDecVector
  end interface

  interface raDecToTan
    module procedure raDecToTanScalar
    module procedure raDecToTanVector
  end interface

  interface detXYtoTan
    module procedure detXYtoTanScalar
    module procedure detXYtoTanVector
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine raDecToDetXY(raDeg, decDeg, scAttitude&
  , myTimeStamp, xDet, yDet, myThetaArcsec, myPhi, status)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(in) :: raDeg(:),&
                                                decDeg(size(raDeg))
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in) :: myTimeStamp
    real(single)     , optional, intent(out) :: xDet(size(raDeg)),&
                                                yDet(size(raDeg))
    real(double)     , optional, intent(out) :: myThetaArcsec(size(raDeg)),&
                                                myPhi(size(raDeg))
    integer,           optional, intent(out) :: status

    type(BoresightT) ::&
      boresight ! typedef from cal/f90/caltypes.f90
    type(EquatorialDirectionType), allocatable ::&
      skyVectors(:) ! typedef from caloalutils/src/caloaldefs.f90
    real(double), dimension(:), pointer ::&
      thetaArcsec, phi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm
    integer(int32) ::&
      numPositions, i
    integer ::&
      localStatus
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Noisy, 'ssclib: ra/dec to XY.')

    localStatus = 0

    if (present(xDet)) xDet = 0.0
    if (present(yDet)) yDet = 0.0
    if (present(myThetaArcsec)) myThetaArcsec = 0.0
    if (present(myPhi)) myPhi = 0.0

    numPositions = size(raDeg)

!    call CAL_setState(instrument=instrumentId, randomize=.false.&
!    , ccdModeId=PRIME_FULL_WINDOW, accuracyLevel=ACCURACY_LOW)
!!*** Need the ccdModeId??

    boresight = CAL_getBoresight(myTimeStamp) ! needs instrument
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm) ! needs instrument
    mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

    allocate(skyVectors(numPositions))
    do i = 1, numPositions
      skyVectors(i)%ra = raDeg(i) * PiOn180
      do while(skyVectors(i)%ra .ge. TwoPi)
        skyVectors(i)%ra = skyVectors(i)%ra - TwoPi
      end do
      do while(skyVectors(i)%ra .lt. 0.0)
        skyVectors(i)%ra = skyVectors(i)%ra + TwoPi
      end do

      if (decDeg(i) .lt. -90.0 .or. decDeg(i) .gt. 90.0) then
        if (present(status)) status = -1
  return
      end if

      skyVectors(i)%dec = min(max(decDeg(i) * PiOn180, -PiOverTwo)&
      , PiOverTwo)
    end do

    call CAL_withSacCoordOrientation(boresight, scAttitude) ! needs instrument
    call CAL_projectOntoTelCoord(boresight, skyVectors, thetaArcsec, phi) ! needs instrument.
    ! NOTE! Units of phi are radians.
    call CAL_telCoordToCamCoord2(thetaArcsec, phi, cc2x, cc2y, cc2z) ! needs only instrument?

    if (present(xDet)) xDet = cc2X * mmToDetUnits
    if (present(yDet)) yDet = cc2Y * mmToDetUnits
    if (present(myThetaArcsec)) myThetaArcsec = thetaArcsec
    if (present(myPhi)) myPhi = phi

    call CAL_releaseMemory(thetaArcsec)
    call CAL_releaseMemory(phi)
    call CAL_releaseMemory(cc2x, cc2y, cc2z)

    deallocate(skyVectors)

    if (present(status)) status = localStatus

  end subroutine raDecToDetXY

  !---------------------------------------------------------------------
  subroutine detXYtoRaDecScalar(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(out) ::&
      raDeg,&
      decDeg
    type(SpacecraftAttitudeType), intent(in)  ::&
      scAttitude ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  ::&
      timeStamp
    real(single),       optional, intent(in)  ::&
      xDet,&
      yDet
    real(double),       optional, intent(in)  ::&
      thetaArcsec,&
      phi
    integer,            optional, intent(out) ::&
      status

    real(single) ::&
      vectorXdet(1),&
      vectorYdet(1)
    real(double) ::&
      vectorRaDeg(1),&
      vectorDecDeg(1),&
      vectorThetaArcsec(1),&
      vectorPhi(1)
    integer ::&
      foreignStatus

    !-------------------------------------------------------------------
    ! Start

    raDeg  = 0.0 ! default
    decDeg = 0.0 ! default

    if (present(xDet) .and. present(yDet)) then
      vectorXdet(1) = xDet
      vectorYdet(1) = yDet

      call detXYtoRaDecVector(vectorRaDeg, vectorDecDeg, scAttitude&
        , timeStamp, vectorXdet, vectorYdet, status=foreignStatus)

    else if (present(thetaArcsec) .and. present(phi)) then
      vectorThetaArcsec(1) = thetaArcsec
      vectorPhi(1)         = phi

      call detXYtoRaDecVector(vectorRaDeg, vectorDecDeg, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi&
        , status=foreignStatus)

    else
      if (present(status)) status = -1
  return
    end if

    raDeg  = vectorRaDeg(1)
    decDeg = vectorDecDeg(1)
    if (present(status)) status = foreignStatus

  end subroutine detXYtoRaDecScalar

  !---------------------------------------------------------------------
  subroutine detXYtoRaDecVector(raDeg, decDeg, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double),                 intent(out) :: raDeg(:),&
                                                 decDeg(size(raDeg))
    type(SpacecraftAttitudeType), intent(in)  :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double),                 intent(in)  :: timeStamp
    real(single),       optional, intent(in)  :: xDet(size(raDeg)),&
                                                 yDet(size(raDeg))
    real(double),       optional, intent(in)  :: thetaArcsec(size(raDeg)),&
                                                 phi(size(raDeg))
    integer,            optional, intent(out) :: status

    real(double), dimension(:), pointer ::&
      localThetaArcsec, localPhi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm, tanTheta, detCartesian(3),&
      skyCartesian(3), transposedInstrumAttitudeMatrix(3,3), ra, dec
    integer(int32) ::&
      numPositions, i
    integer ::&
      localStatus
    type(SpacecraftAttitudeType) ::&
      instrumAttitude

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    raDeg  = 0.0 ! default
    decDeg = 0.0 ! default

    numPositions = size(raDeg)

    if (present(xDet) .and. present(yDet)) then
      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

      allocate(cc2X(numPositions), cc2Y(numPositions))
      cc2X = xDet / mmToDetUnits
      cc2Y = yDet / mmToDetUnits
      cc2z = focalLengthMm

      call CAL_camCoord2ToTelCoord(cc2x, cc2y, cc2z, localThetaArcsec, localPhi)

      deallocate(cc2X, cc2Y, cc2z)

    else if (present(thetaArcsec) .and. present(phi)) then
      allocate(localThetaArcsec(numPositions), localPhi(numPositions))
      localThetaArcsec = thetaArcsec
      localPhi = phi

    else
      if (present(status)) status = -1
  return
    end if

    instrumAttitude = getInstrumentAttitude(scAttitude, timeStamp)

    call OAL_toAttitudeMatrix(instrumAttitude, transposedInstrumAttitudeMatrix)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_in = AIN_sky * v_sky.
    ! or
    !		v_sky = AIN_sky^-1 * v_in.
    !
    ! Note that the OAL call returns transpose(AIN_sky) = AIN_sky^-1.

    do i = 1, numPositions
      tanTheta = tan(localThetaArcsec(i) * ArcsecToRad)
      detCartesian(1) = -1.0
      detCartesian(2) = tanTheta * sin(localPhi(i))
      detCartesian(3) = tanTheta * cos(localPhi(i))

      skyCartesian = matmul(transposedInstrumAttitudeMatrix, detCartesian)
      call cartesianToSky(skyCartesian, ra, dec)
      raDeg(i)  = ra  / PiOn180
      decDeg(i) = dec / PiOn180
    end do

    if (present(xDet) .and. present(yDet)) then
      call CAL_releaseMemory(localThetaArcsec)
      call CAL_releaseMemory(localPhi)

    else if (present(thetaArcsec) .and. present(phi)) then
      deallocate(localThetaArcsec, localPhi)
    end if

!**** other CAL_releaseMemory

    if (present(status)) status = localStatus

  end subroutine detXYtoRaDecVector

  !---------------------------------------------------------------------
  subroutine detXYtoTanScalar(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(out)           :: tanX, tanY
    real(double), intent(in)            :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double), intent(in)            :: timeStamp
    real(single), intent(in),  optional :: xDet,&
                                           yDet
    real(double), intent(in),  optional :: thetaArcsec,&
                                           phi
    integer,      intent(out), optional :: status

    real(single) ::&
      vectorXdet(1), vectorYdet(1)
    real(double) ::&
      vectorTanX(1), vectorTanY(1),&
      vectorThetaArcsec(1), vectorPhi(1)
    integer ::&
      foreignStatus

    !-------------------------------------------------------------------
    ! Start

    tanX = 0.0 ! default
    tanY = 0.0 ! default

    if (present(xDet) .and. present(yDet)) then
      vectorXdet(1) = xDet
      vectorYdet(1) = yDet

      call detXYtoTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, vectorXdet, vectorYdet, status=foreignStatus)

    else if (present(thetaArcsec) .and. present(phi)) then
      vectorThetaArcsec(1) = thetaArcsec
      vectorPhi(1)         = phi

      call detXYtoTanVector(vectorTanX, vectorTanY, refRa, refDec, scAttitude&
        , timeStamp, thetaArcsec=vectorThetaArcsec, phi=vectorPhi&
        , status=foreignStatus)

    else
      if (present(status)) status = -1
  return
    end if

    tanX = vectorTanX(1)
    tanY = vectorTanY(1)
    if (present(status)) status = foreignStatus

  end subroutine detXYtoTanScalar

  !---------------------------------------------------------------------
  subroutine detXYtoTanVector(tanX, tanY, refRa, refDec, scAttitude&
  , timeStamp, xDet, yDet, thetaArcsec, phi, status)
  !---------------------------------------------------------------------
  ! NOTE! Cal must be set up for the correct instrument by the calling
  ! routine.

    implicit none

    real(double), intent(out)           :: tanX(:),&
                                           tanY(size(tanX))
    real(double), intent(in)            :: refRa, refDec
    type(SpacecraftAttitudeType), intent(in) :: scAttitude
    ! typedef from caloalutils/src/caloaldefs.f90
    real(double), intent(in)            :: timeStamp
    real(single), intent(in),  optional :: xDet(size(tanX)),&
                                           yDet(size(tanX))
    real(double), intent(in),  optional :: thetaArcsec(size(tanX)),&
                                           phi(size(tanX))
    integer,      intent(out), optional :: status

    real(double), dimension(:), pointer ::&
      localThetaArcsec, localPhi, cc2x, cc2y, cc2z
    real(double) ::&
      mmToDetUnits, focalLengthMm, tanTheta,&
      srcVector_ins(3), srcVector_ref(3),&
      transpInstrumMatrix_sky(3,3), refMatrix_sky(3,3), refMatrix_ins(3,3)
    integer(int32) ::&
      numPositions, i
    integer ::&
      localStatus
    type(SpacecraftAttitudeType) ::&
      instrumAttitude
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    tanX = 0.0 ! default
    tanY = 0.0 ! default

    numPositions = size(tanX)

    if (present(xDet) .and. present(yDet)) then
      call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
      mmToDetUnits = RadToArcsec / focalLengthMm / detUnitArcsec

      allocate(cc2X(numPositions), cc2Y(numPositions))
      cc2X = xDet / mmToDetUnits
      cc2Y = yDet / mmToDetUnits
      cc2z = focalLengthMm

      call CAL_camCoord2ToTelCoord(cc2x, cc2y, cc2z, localThetaArcsec, localPhi)

      deallocate(cc2X, cc2Y, cc2z)

    else if (present(thetaArcsec) .and. present(phi)) then
      allocate(localThetaArcsec(numPositions), localPhi(numPositions))
      localThetaArcsec = thetaArcsec
      localPhi = phi

    else
      if (present(status)) status = -1
  return
    end if

    instrumAttitude = getInstrumentAttitude(scAttitude, timeStamp)
if (testFlag) print *, 'instrumAttitude%raDeg, instrumAttitude%decDeg: '&
, instrumAttitude%ra / PiOn180, instrumAttitude%dec / PiOn180

    call OAL_toAttitudeMatrix(instrumAttitude, transpInstrumMatrix_sky)
    !
    ! The array returned by this call represents the instrument attitude
    ! matrix AIN_sky which has, in its ith row, a vector of direction
    ! cosines in the sky basis of the ith axis of the instrument
    ! reference frame (the same frame as defines the TELCOORD system).
    ! For example, the x or 1st axis of this frame appears in the first row of
    ! AIN_sky. The sky basis is that right-handed frame which has its
    ! x and y axes at declination=0, z axis pointing in the declination
    ! = pi/2 direction, x in the ra=0 and y in the ra=pi/2 (ie, 6 hrs)
    ! directions.
    !
    ! This matrix can be used to convert a vector from the sky basis to
    ! the instrument or TELCOORD basis, as follows:
    !
    !		v_in = AIN_sky * v_sky.
    ! or
    !		v_sky = AIN_sky^-1 * v_in.
    !
    ! Note that the OAL call in fact returns transpose(AIN_sky) = AIN_sky^-1.

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :), refMatrix_sky(2, :))

    refMatrix_ins = matmul(refMatrix_sky, transpInstrumMatrix_sky)
!*** the (1,1) number may be inaccurate - should use a better routine.

    do i = 1, numPositions
      tanTheta = tan(localThetaArcsec(i) * ArcsecToRad)
      srcVector_ins(1) = -1.0
      srcVector_ins(2) = tanTheta * sin(localPhi(i))
      srcVector_ins(3) = tanTheta * cos(localPhi(i))

      srcVector_ref = matmul(refMatrix_ins, srcVector_ins)

      tanX(i) = -srcVector_ref(2)
      tanY(i) = srcVector_ref(3)
    end do

    if (present(xDet) .and. present(yDet)) then
      call CAL_releaseMemory(localThetaArcsec)
      call CAL_releaseMemory(localPhi)

    else if (present(thetaArcsec) .and. present(phi)) then
      deallocate(localThetaArcsec, localPhi)
    end if

!**** other CAL_releaseMemory

    if (present(status)) status = localStatus

  end subroutine detXYtoTanVector

  !---------------------------------------------------------------------
  function toAttitudePatch(eulerAngles) result(attitude)
    !---------------------------------------------------------------------
    type(EulerAnglesType), intent(in)   ::&
      eulerAngles ! typedef from oal/f90/oaltypes.f90
    type(SpacecraftAttitudeType)        ::&
      attitude ! typedef from caloalutils/src/caloaldefs.f90

    attitude%ra   =  eulerAngles%phi 
    attitude%dec  = -eulerAngles%theta
    attitude%apos = -eulerAngles%psi

    if (attitude%ra <0) attitude%ra = attitude%ra + 2*pi     ! 0 < ra < 2*pi
    if (attitude%apos <0) attitude%apos = attitude%apos + 2*pi     ! 0 < apos < 2*pi


  end function toAttitudePatch

  !---------------------------------------------------------------------
  function toEulerAngles(matrix) result(eulerAngles)
    !---------------------------------------------------------------------

    real(double), dimension(3,3), intent(in)  :: matrix
    type(EulerAnglesType)                     ::&
      eulerAngles ! typedef from oal/f90/oaltypes.f90

    ! need to transpose matrix because OAL (c++) and F90 store matrices transposed.
    call OAL_toEulerAngles(transpose(matrix), eulerAngles)

  end function toEulerAngles

  !---------------------------------------------------------------------
  function toMatrix(attitude) result(matrix)
    !---------------------------------------------------------------------
    type(SpacecraftAttitudeType), intent(in) ::&
      attitude ! typedef from caloalutils/src/caloaldefs.f90
    real(double), dimension(3,3)             :: matrix

    call OAL_toAttitudeMatrix(attitude, matrix)

    ! need to transpose matrix because OAL (c++) and F90 store matrices transposed.
    matrix=transpose(matrix)

  end function toMatrix

  !---------------------------------------------------------------------
  subroutine raDecToTanScalar(refRa, refDec, ra, dec, tanX, tanY, status)
  !---------------------------------------------------------------------
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)  :: refRa, refDec, ra, dec
    real(double), intent(out) :: tanX, tanY
    integer, intent(out), optional :: status

    real(double) :: localRa(1), localDec(1), localTanX(1), localTanY(1)
    integer :: localStatus

    !-------------------------------------------------------------------
    ! Start

    localRa(1) = ra
    localDec(1) = dec

    call raDecToTanVector(refRa, refDec, localRa, localDec&
    , localTanX, localTanY, localStatus)

    tanX = localTanX(1)
    tanY = localTanY(1)

    if (present(status)) status = localStatus

  end subroutine raDecToTanScalar

  !---------------------------------------------------------------------
  subroutine raDecToTanVector(refRa, refDec, ra, dec, tanX, tanY, status)
  !---------------------------------------------------------------------
  ! NOTE that all ras and decs must be in RADIANS!!

use dump_mod	! contains printMatrix

    implicit none

    real(double), intent(in)  :: refRa, refDec, ra(:), dec(:)
    real(double), intent(out) :: tanX(:), tanY(:)
    integer, intent(out), optional :: status

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3)
    integer(int32) :: vsize, i
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) then
print *, 'vvvvvvv'
print *, 'Start of raDecToTan'
end if

    if (present(status)) status = 0

    tanX  = 0.0 ! default.
    tanY = 0.0 ! default.

    vsize = size(ra)
    if (size(dec) .ne. vsize .or. size(tanX) .ne. vsize&
    .or. size(tanY) .ne. vsize) then
      if (present(status)) status = -1
  return
    end if
    if (vsize .lt. 1) then
      if (present(status)) status = -2
  return
    end if

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :), refMatrix_sky(2, :))

if (testFlag) then
print *, 'refRa, refDec: ', refRa / PiOn180, refDec / PiOn180
print *, ''
call printMatrix('The sky ref matrix:', refMatrix_sky)
print *, ''
end if
    do i = 1, vsize
      srcVector_sky = skyToCartesian(ra(i), dec(i))
      srcVector_ref = matmul(refMatrix_sky, srcVector_sky)

      tanX(i) = -srcVector_ref(2)
      tanY(i) =  srcVector_ref(3)
if (testFlag) then
print *, '  i: ', i
print *, '  srcVector_sky: ', srcVector_sky
print *, '  srcVector_ref: ', srcVector_ref
print *, '  tanX(i) (= -srcVector_ref(2)): ', tanX(i)
print *, '  tanY(i) (=  srcVector_ref(3)): ', tanY(i)
end if
    end do

if (testFlag) then
print *, 'End of raDecToTan'
print *, '^^^^^^^'
end if
  end subroutine raDecToTanVector

  !---------------------------------------------------------------------
  subroutine tanToRaDec(refRa, refDec, tanX, tanY, ra, dec, status)
  !---------------------------------------------------------------------
  ! This subroutine converts positions from a tangent-plane coordinate system into ra and dec. The x-axis of the tangent plane points to the right and is in the direction of decreasing ra; the y-axis points up and is in the direction of increasing dec.
  !
  ! NOTE that all ras and decs must be in RADIANS!!

    implicit none

    real(double), intent(in)  :: refRa, refDec, tanX(:), tanY(:)
    real(double), intent(out) :: ra(:), dec(:)
    integer, intent(out), optional :: status

    real(double) :: srcVector_sky(3), refMatrix_sky(3,3), srcVector_ref(3),&
      norm
    integer(int32) :: vsize, i

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0

    ra  = 0.0 ! default.
    dec = 0.0 ! default.

    vsize = size(tanX)
    if (size(tanY) .ne. vsize .or. size(ra) .ne. vsize&
    .or. size(dec) .ne. vsize) then
      if (present(status)) status = -1
  return
    end if
    if (vsize .lt. 1) then
      if (present(status)) status = -2
  return
    end if

    refMatrix_sky(1, :) = skyToCartesian(refRa, refDec)
    refMatrix_sky(2, :) = skyToCartesian(refRa + PiOverTwo, 0.0d0)
    refMatrix_sky(3, :) = vectorCross(refMatrix_sky(1, :)&
    , refMatrix_sky(2, :))
!*** how about when ref is pointing straight up or straight down.

    do i = 1, vsize
      norm = sqrt(1.0 + tanX(i) * tanX(i) + tanY(i) * tanY(i))
      srcVector_ref(1) = 1.0 / norm
      srcVector_ref(2) = -tanX(i) / norm
      srcVector_ref(3) = tanY(i) / norm

      srcVector_sky = matmul(srcVector_ref, refMatrix_sky)
      call cartesianToSky(srcVector_sky, ra(i), dec(i))
    end do

  end subroutine tanToRaDec

  !---------------------------------------------------------------------
  function skyToCartesian(ra, dec) result(vector)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in) :: ra, dec 
    real(double)             :: vector(3)

    real (double) :: cosDec 

    !-------------------------------------------------------------------
    ! Start

    cosDec = cos(dec)
    vector(1) = cos(ra) * cosDec
    vector(2) = sin(ra) * cosDec
    vector(3) = sin(dec)

  end function skyToCartesian

  !---------------------------------------------------------------------
  subroutine cartesianToSky(vector, ra, dec)
  !---------------------------------------------------------------------

    implicit none

    real(double), intent(in)  :: vector(3) 
    real(double), intent(out) :: ra, dec

    !-------------------------------------------------------------------
    ! Start

    if (vector(1) .ne. 0.0 .and. vector(2) .ne. 0.0) then
      ra = atan2(vector(2), vector(1))
      do while(ra .lt. 0.0)
        ra = ra + TwoPi
      end do
    else
      ra = 0.0
    end if
    dec = asin(vector(3))

  end subroutine cartesianToSky

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module coordinate
