!
! Module: fits_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare aa ptrs => null()!
! 
! $Id: fits_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module fits_utils

  use dal

  use sscmisc
!  use get_free_io_unit_mod

  implicit none 

  private 

  public getRealKwd
  public getIntKwd
  public getStringKwd
  public copyAttribute
  public readColData
  public myAddColumn
  public minNonNullValue
  public maxNonNullValue
  public loadFitsArray
  public mergeTables
  public saveOutImage
  public copyVitalArrayKwds

  interface copyAttribute
    module procedure copyAttribute_arrayToArray
    module procedure copyAttribute_arrayToSet
    module procedure copyAttribute_setToSet
    module procedure copyAttributeTabToTab
  end interface

  interface readColData
    module procedure readColData_real32_wn
    module procedure readColData_real64_wn
    module procedure readColData_int8_wn
    module procedure readColData_int16_wn
    module procedure readColData_int32_wn
    module procedure readColData_real32
    module procedure readColData_real64
    module procedure readColData_int8
    module procedure readColData_int16
    module procedure readColData_int32
  end interface

  interface myAddColumn
    module procedure myAddColumn_int32
    module procedure myAddColumn_real32
    module procedure myAddColumn_real64
  end interface

  interface minNonNullValue
    module procedure minNonNullValue_real32
    module procedure minNonNullValue_real64
  end interface

  interface maxNonNullValue
    module procedure maxNonNullValue_real32
    module procedure maxNonNullValue_real64
  end interface

  interface loadFitsArray
    module procedure loadFitsArray_1d_r32, loadFitsArray_1d_wa_r32
    module procedure loadFitsArray_2d_r32, loadFitsArray_2d_wa_r32
    module procedure loadFitsArray_3d_r32, loadFitsArray_3d_wa_r32
    module procedure loadFitsArray_4d_r32, loadFitsArray_4d_wa_r32
    module procedure loadFitsArray_1d_r64, loadFitsArray_1d_wa_r64
    module procedure loadFitsArray_2d_r64, loadFitsArray_2d_wa_r64
    module procedure loadFitsArray_3d_r64, loadFitsArray_3d_wa_r64
    module procedure loadFitsArray_4d_r64, loadFitsArray_4d_wa_r64
    module procedure loadFitsArray2dSingle
    module procedure loadFitsArray2dBoolean
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine getRealKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS real32 or 64 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    real(double),   intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0.0d0 ! default

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. REAL32) then
        value = real32Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. REAL64) then
          value = real64Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. REAL32) then
        value = real32Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. REAL64) then
          value = real64Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getRealKwd

  !---------------------------------------------------------------------
  subroutine getIntKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS int8, 16 or 32 attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

    implicit none

    character*(*),  intent(in)            :: keyword
    integer(int32), intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = 0 ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER8) then
        value = int8Attribute(inDataSet, keyword)
      else if (dataType(attribute(inDataSet, keyword)) .eq. INTEGER16) then
        value = int16Attribute(inDataSet, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(inDataSet, keyword)) .eq. INTEGER32) then
          value = int32Attribute(inDataSet, keyword)
        else
          localStatus = -1
        end if
      end if
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      if (dataType(attribute(imageArray, keyword)) .eq. INTEGER8) then
        value = int8Attribute(imageArray, keyword)
      else if (dataType(attribute(imageArray, keyword)) .eq. INTEGER16) then
        value = int16Attribute(imageArray, keyword)
      else
        if (fatalFail&
        .or. dataType(attribute(imageArray, keyword)) .eq. INTEGER32) then
          value = int32Attribute(imageArray, keyword)
        else
          localStatus = -2
        end if
      end if
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getIntKwd

  !---------------------------------------------------------------------
  subroutine getStringKwd(keyword, inDataSet, imageArray, value, status)
  !---------------------------------------------------------------------
  ! Gets FITS string attribute from main dataset or primary array.
  !
  ! These routines were adapted from Clive Page's implot.

!****Here it is assumed that a keyword of any type can be accessed as string?? Should check this.

    implicit none

    character*(*),  intent(in)            :: keyword
    character*(*),  intent(out)           :: value
    integer,        intent(out), optional :: status
    type(DataSetT), intent(in)            :: inDataSet
    type(ArrayT),   intent(in)            :: imageArray

    integer :: localStatus
    logical :: fatalFail

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0
    value = "??" ! default.

    if (present(status)) then
      fatalFail = .false.
    else
      fatalFail = .true.
    end if

    if(hasAttribute(inDataSet, keyword)) then
      value = stringAttribute(inDataSet, keyword)
    else if (fatalFail .or. hasAttribute(imageArray, keyword)) then
      value = stringAttribute(imageArray, keyword)
    else
      localStatus = -3
    end if

    if (present(status)) status = localStatus

  end subroutine getStringKwd

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToArray(inArray, kwdName, outArray, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outArray, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToArray

  !---------------------------------------------------------------------
  subroutine copyAttribute_arrayToSet(inArray, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    type(DataSetT), intent(in) :: outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inArray, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inArray, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_arrayToSet

  !---------------------------------------------------------------------
  subroutine copyAttribute_setToSet(inSet, kwdName, outSet, status)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: inSet, outSet
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inSet, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outSet, attribute(inSet, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttribute_setToSet

  !---------------------------------------------------------------------
  subroutine copyAttributeTabToTab(inTab, kwdName, outTab, status)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: inTab, outTab
    character(*), intent(in) :: kwdName
    integer, intent(out), optional :: status

    !-------------------------------------------------------------------
    ! Start

    if (present(status) .and. .not. hasAttribute(inTab, kwdName)) then
      status = -1
  return
    end if

    call setAttribute(outTab, attribute(inTab, kwdName))
    if (present(status)) status = 0

  end subroutine copyAttributeTabToTab

  !---------------------------------------------------------------------
  subroutine readColData_real32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real64_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_real64_wn

  !---------------------------------------------------------------------
  subroutine readColData_int8_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int8_wn

  !---------------------------------------------------------------------
  subroutine readColData_int16_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int16_wn

  !---------------------------------------------------------------------
  subroutine readColData_int32_wn(tab, colName, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(TableT), intent(in)            :: tab
    character(*), intent(in)            :: colName
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    type(ColumnT) :: col
    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(tab)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))
    col = column(tab, colName)

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    call release(col)

    if (present(status)) status = localStatus

  end subroutine readColData_int32_wn

  !---------------------------------------------------------------------
  subroutine readColData_real32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(single), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, single)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, single)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, single)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = colDataReal32
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = real(colDataReal64, single)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real32

  !---------------------------------------------------------------------
  subroutine readColData_real64(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    real(double), intent(out)           :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0.0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_real64

  !---------------------------------------------------------------------
  subroutine readColData_int8(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int8 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int8), intent(out)          :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = colDataInt8
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int8)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = int(colDataInt32, int8)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int8)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int8)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int8

  !---------------------------------------------------------------------
  subroutine readColData_int16(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int16 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int16), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int16)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int16)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int16)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int16)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int16

  !---------------------------------------------------------------------
  subroutine readColData_int32(col, colData, status)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in an int32 vector colData.

    implicit none 

    type(ColumnT), intent(in)           :: col
    integer(int32), intent(out)         :: colData(:)
    integer,      intent(out), optional :: status

    integer :: localStatus
    integer(int8), dimension(:), pointer ::&
      colDataInt8
    integer(int16), dimension(:), pointer ::&
      colDataInt16
    integer(int32), dimension(:), pointer ::&
      colDataInt32
    real(single), dimension(:), pointer ::&
      colDataReal32
    real(double), dimension(:), pointer ::&
      colDataReal64

    !-------------------------------------------------------------------
    ! Start:

    localStatus = 0

    colData = 0
    if (size(colData) .ne. numberOfRows(col)) then
      if (present(status)) status = -1
  return
    end if

!    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = int(colDataInt8, int32)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = int(colDataInt16, int32)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = colDataInt32
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = int(colDataReal32, int32)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = int(colDataReal64, int32)
    end select

    if (present(status)) status = localStatus

  end subroutine readColData_int32

  !---------------------------------------------------------------------
  subroutine myAddColumn_int32(tab, name, inData, unit, comment, nulls&
  , nullValue)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    integer(int32), intent(in), optional :: nullValue
    integer(int32), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    integer(int32), pointer :: colData(:)
    integer(int32) :: i, localNullValue

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nullValue)) then
      localNullValue = nullValue
    else
      localNullValue = -999
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, INTEGER32, trim(localUnit)&
    , comment=trim(localComment))

    colData => int32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      call setNullValue(col, localNullValue)
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_int32

  !---------------------------------------------------------------------
  subroutine myAddColumn_real32(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(single), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(single), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL32, trim(localUnit)&
    , comment=trim(localComment))

    colData => real32Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_real32

  !---------------------------------------------------------------------
  subroutine myAddColumn_real64(tab, name, inData, unit, comment, nulls)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: name
    character(*), intent(in), optional :: unit, comment
    logical, intent(in), optional :: nulls(:)
    real(double), intent(in) :: inData(:)

    character(255) :: localUnit, localComment
    logical :: localNulls(size(inData))
    type(ColumnT) :: col
    real(double), pointer :: colData(:)
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start

    if (present(unit)) then
      localUnit = unit
    else
      localUnit = ''
    end if
    if (present(comment)) then
      localComment = comment
    else
      localComment = ''
    end if
    if (present(nulls)) then
      localNulls = .true.
      if (size(nulls) .eq. size(inData)) then
        localNulls = nulls
      else if (size(nulls) .gt. size(inData)) then
        localNulls = nulls(1:size(inData))
      else
        localNulls(1:size(nulls)) = nulls
      end if
    else
      localNulls = .false.
    end if

    col = addColumn(tab, name, REAL64, trim(localUnit)&
    , comment=trim(localComment))

    colData => real64Data(col)
    colData = inData(1:size(inData))

    if (any(localNulls)) then
      do i = 1, size(inData)
        if (localNulls(i)) call setToNull(col, i - 1)
      end do
    end if

    call release(col)

  end subroutine myAddColumn_real64

  !---------------------------------------------------------------------
  subroutine minNonNullValue_real32(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), intent(out) :: minValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValue_real32

  !---------------------------------------------------------------------
  subroutine minNonNullValue_real64(tab, colName, minValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), intent(out) :: minValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    minValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    minValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        minValue = min(minValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine minNonNullValue_real64

  !---------------------------------------------------------------------
  subroutine maxNonNullValue_real32(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(single), intent(out) :: maxValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(single), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real32Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValue_real32

  !---------------------------------------------------------------------
  subroutine maxNonNullValue_real64(tab, colName, maxValue, allRowsNull)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: tab
    character(*), intent(in) :: colName
    real(double), intent(out) :: maxValue
    logical, intent(out) :: allRowsNull

    type(ColumnT) :: col
    integer(int32) :: firstNonNullRow, row
    real(double), pointer :: colData(:)

    !-------------------------------------------------------------------
    ! Start

    maxValue = 0.0

    col = column(tab, colName)
    allRowsNull = .true. ! default
    firstNonNullRow = -1
    do row = 0, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        allRowsNull = .false.
        firstNonNullRow = row
    exit
      end if
    end do

    if (allRowsNull) then
      call release(col)
  return
    end if

    colData => real64Data(col)
    maxValue = colData(firstNonNullRow)
    do row = firstNonNullRow + 1, numberOfRows(col) - 1
      if (.not. isNull(col, row)) then
        maxValue = max(maxValue, colData(row))
      end if
    end do

    call release(col)

  end subroutine maxNonNullValue_real64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_r32(imageSetName, vector)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:), pointer :: vector

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_1d_wa_r32(inArray, vector)!!!, dataType)

  end subroutine loadFitsArray_1d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_r32(imageSetName, image)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:), pointer :: image

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)

  end subroutine loadFitsArray_2d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_r32(imageSetName, cube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:,:), pointer :: cube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)

  end subroutine loadFitsArray_3d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_r32(imageSetName, hyperCube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(single), dimension(:,:,:,:), pointer :: hyperCube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)

  end subroutine loadFitsArray_4d_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_r64(imageSetName, vector)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:), pointer :: vector

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_1d_wa_r64(inArray, vector)!!!, dataType)

  end subroutine loadFitsArray_1d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_r64(imageSetName, image)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:), pointer :: image

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)

  end subroutine loadFitsArray_2d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_r64(imageSetName, cube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:,:), pointer :: cube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)

  end subroutine loadFitsArray_3d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_r64(imageSetName, hyperCube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: imageSetName
    real(double), dimension(:,:,:,:), pointer :: hyperCube

    type(DataSetT) :: set
    type(ArrayT) :: inArray

    !-------------------------------------------------------------------
    ! Start:

    set = dataSet(imageSetName, READ)
    inArray = array(set, 0, READ)
    call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)

  end subroutine loadFitsArray_4d_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_wa_r32(inArray, vector)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate vector!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:), pointer :: vector

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:), pointer :: image
    real(single), dimension(:,:,:), pointer :: cube
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)

    case(3)
      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badHyperCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(hyperCube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(hyperCube, vectorShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_1d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_wa_r32(inArray, image)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate image!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:,:), pointer :: cube
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = real(arrayDataReal64, single)
      end if

    case(3)
      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_2d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate cube!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:,:), pointer :: cube

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(3, 1), cubeShape(3)
    logical :: mask(4)

    integer(int8), dimension(:,:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = real(arrayDataReal64, single)
      end if

    case(4)
      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 3) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      mask(biggestDimensions(2,1)) = .false.
      biggestDimensions(3,:) = maxloc(arrayShape, mask)
      cubeShape(1) = size(hyperCube, biggestDimensions(1,1))
      cubeShape(2) = size(hyperCube, biggestDimensions(2,1))
      cubeShape(3) = size(hyperCube, biggestDimensions(3,1))
      allocate(cube(cubeShape(1), cubeShape(2), cubeShape(3)))
      cube = reshape(hyperCube, cubeShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_3d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_wa_r32(inArray, hypercube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(single), dimension(:,:,:,:), pointer :: hyperCube

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer :: numDimensions

!    integer(int8),  dimension(:,:,:,:), pointer ::&
!      arrayDataInt8
!    integer(int16), dimension(:,:,:,:), pointer ::&
!      arrayDataInt16
!    integer(int32), dimension(:,:,:,:), pointer ::&
!      arrayDataInt32
!    real(single),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal32
!    real(double),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2, 3)
call error('arrayRankTooSmall', '')
    case(4)
call error('notAvailableYet', "Reading of 4-dimensional arrays doesn't &
&seem to be available in the DAL yet.")

!      allocate(hypercube(dimVector(0), dimVector(1), dimVector(2)&
!      , dimVector(3)))
!
!      if (dataType(inArray) .eq. INTEGER8) then
!        arrayDataInt8    => int8Array4Data(inArray)
!        hyperCube = real(arrayDataInt8, double)
!      else if (dataType(inArray) .eq. INTEGER16) then
!        arrayDataInt16   => int16Array4Data(inArray)
!        hyperCube = real(arrayDataInt16, double)
!      else if (dataType(inArray) .eq. INTEGER32) then
!        arrayDataInt32   => int32Array4Data(inArray)
!        hyperCube = real(arrayDataInt32, double)
!      else if (dataType(inArray) .eq. REAL32) then
!        arrayDataReal32 => real32Array4Data(inArray)
!        hyperCube = real(arrayDataReal32, double)
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array4Data(inArray)
!        hyperCube = arrayDataReal64
!      end if
hyperCube = 0.0 ! keeps the compiler from issuing a warning.

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_4d_wa_r32

  !---------------------------------------------------------------------
  subroutine loadFitsArray_1d_wa_r64(inArray, vector)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate vector!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:), pointer :: vector

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:), pointer :: image
    real(double), dimension(:,:,:), pointer :: cube
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimension(1), vectorShape(1)

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      call loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)
      arrayShape(:2) = shape(image)
if (count(arrayShape .gt. 1) .gt. 1) call error('badImageShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(image, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(image, vectorShape)

    case(3)
      call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(cube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(cube, vectorShape)

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 1) call error('badHyperCubeShape','')
      biggestDimension = maxloc(arrayShape)
      vectorShape(1) = size(hyperCube, biggestDimension(1))
      allocate(vector(vectorShape(1)))
      vector = reshape(hyperCube, vectorShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_1d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_2d_wa_r64(inArray, image)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate image!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:), pointer :: image

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:,:), pointer :: cube
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)

    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(image(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        image = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        image = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        image = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        image = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        image = arrayDataReal64
      end if

    case(3)
      call loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
      arrayShape(:3) = shape(cube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(cube, imageShape)

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      imageShape(1) = size(cube, biggestDimensions(1,1))
      imageShape(2) = size(cube, biggestDimensions(2,1))
      allocate(image(imageShape(1), imageShape(2)))
      image = reshape(hyperCube, imageShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_2d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_3d_wa_r64(inArray, cube)!!!, dataType)
  !---------------------------------------------------------------------
!**** may need to deallocate cube!!

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:,:), pointer :: cube

    integer(int32), dimension(:), pointer ::&
      dimVector
    real(double), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4), biggestDimensions(3, 1), cubeShape(3)
    logical :: mask(4)

    integer(int8), dimension(:,:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1, 2)
call error('arrayRankTooSmall', '')
    case(3)
      allocate(cube(dimVector(0), dimVector(1), dimVector(2)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array3Data(inArray)
        cube = real(arrayDataInt8, double)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array3Data(inArray)
        cube = real(arrayDataInt16, double)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array3Data(inArray)
        cube = real(arrayDataInt32, double)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array3Data(inArray)
        cube = real(arrayDataReal32, double)
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array3Data(inArray)
        cube = arrayDataReal64
      end if

    case(4)
      call loadFitsArray_4d_wa_r64(inArray, hyperCube)!!!, dataType)
      arrayShape = shape(hyperCube)
if (count(arrayShape .gt. 1) .gt. 3) call error('badHyperCubeShape','')
      biggestDimensions(1,:) = maxloc(arrayShape)
      mask(biggestDimensions(1,1)) = .false.
      biggestDimensions(2,:) = maxloc(arrayShape, mask)
      mask(biggestDimensions(2,1)) = .false.
      biggestDimensions(3,:) = maxloc(arrayShape, mask)
      cubeShape(1) = size(hyperCube, biggestDimensions(1,1))
      cubeShape(2) = size(hyperCube, biggestDimensions(2,1))
      cubeShape(3) = size(hyperCube, biggestDimensions(3,1))
      allocate(cube(cubeShape(1), cubeShape(2), cubeShape(3)))
      cube = reshape(hyperCube, cubeShape)

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_3d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray_4d_wa_r64(inArray, hypercube)!!!, dataType)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray
    real(double), dimension(:,:,:,:), pointer :: hyperCube

    integer(int32), dimension(:), pointer ::&
      dimVector
    integer :: numDimensions

!    integer(int8),  dimension(:,:,:,:), pointer ::&
!      arrayDataInt8
!    integer(int16), dimension(:,:,:,:), pointer ::&
!      arrayDataInt16
!    integer(int32), dimension(:,:,:,:), pointer ::&
!      arrayDataInt32
!    real(single),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal32
!    real(double),   dimension(:,:,:,:), pointer ::&
!      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    dimVector => dimensions(inArray)

    numDimensions = size(dimVector)

    select case(numDimensions)
    case(1, 2, 3)
call error('arrayRankTooSmall', '')
    case(4)
call error('notAvailableYet', "Reading of 4-dimensional arrays doesn't &
&seem to be available in the DAL yet.")

!      allocate(hypercube(dimVector(0), dimVector(1), dimVector(2)&
!      , dimVector(3)))
!
!      if (dataType(inArray) .eq. INTEGER8) then
!        arrayDataInt8    => int8Array4Data(inArray)
!        hyperCube = real(arrayDataInt8, double)
!      else if (dataType(inArray) .eq. INTEGER16) then
!        arrayDataInt16   => int16Array4Data(inArray)
!        hyperCube = real(arrayDataInt16, double)
!      else if (dataType(inArray) .eq. INTEGER32) then
!        arrayDataInt32   => int32Array4Data(inArray)
!        hyperCube = real(arrayDataInt32, double)
!      else if (dataType(inArray) .eq. REAL32) then
!        arrayDataReal32 => real32Array4Data(inArray)
!        hyperCube = real(arrayDataReal32, double)
!      else ! this gives a DAL error if data type is none of the above.
!        arrayDataReal64 => real64Array4Data(inArray)
!        hyperCube = arrayDataReal64
!      end if
hyperCube = 0.0 ! keeps the compiler from issuing a warning.

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray_4d_wa_r64

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dSingle(setName, blockNumber, fitsArray)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: setName
    integer, intent(in), optional :: blockNumber
    real(single), pointer :: fitsArray(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer :: localBlockNumber
    integer(int32), dimension(:), pointer ::&
      dimVector
!!!    real(single), dimension(:,:,:), pointer :: cube
!!!    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4)
!!!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start

    if (present(blockNumber)) then
      localBlockNumber = blockNumber
    else
      localBlockNumber = 0
    end if

    set = dataSet(setName, READ)
    inArray = array(set, blockNumber, READ)
    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(fitsArray(dimVector(0), dimVector(1)))

      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        fitsArray = real(arrayDataInt8, single)
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        fitsArray = real(arrayDataInt16, single)
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        fitsArray = real(arrayDataInt32, single)
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        fitsArray = arrayDataReal32
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        fitsArray = real(arrayDataReal64, single)
      end if

    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(cube, imageShape)
call error('notSupported', '')

    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(hyperCube, imageShape)
call error('notSupported', '')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray2dSingle

  !---------------------------------------------------------------------
  subroutine loadFitsArray2dBoolean(setName, blockNumber, fitsArray)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: setName
    integer, intent(in), optional :: blockNumber
    logical, pointer :: fitsArray(:,:)

    type(DataSetT) :: set
    type(ArrayT) :: inArray
    integer :: localBlockNumber
    integer(int32), dimension(:), pointer ::&
      dimVector
!!!    real(single), dimension(:,:,:), pointer :: cube
!!!    real(single), dimension(:,:,:,:), pointer :: hyperCube
    integer :: numDimensions, arrayShape(4)
!!!, biggestDimensions(2, 1), imageShape(2)
    logical :: mask(4)
    integer(int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(double), dimension(:,:), pointer ::&
      arrayDataReal64
!    logical(bool), dimension(:,:), pointer ::&
!      arrayDataBoolean

    !-------------------------------------------------------------------
    ! Start

    if (present(blockNumber)) then
      localBlockNumber = blockNumber
    else
      localBlockNumber = 0
    end if

    set = dataSet(setName, READ)
    inArray = array(set, blockNumber, READ)
    dimVector => dimensions(inArray)
    numDimensions = size(dimVector)
    arrayShape = 0 ! vector op, default
    mask = .true.  ! vector op, default

    select case(numDimensions)
    case(1)
call error('arrayRankTooSmall', '')
    case(2)
      allocate(fitsArray(dimVector(0), dimVector(1)))

      fitsArray = .false.
      if (dataType(inArray) .eq. INTEGER8) then
        arrayDataInt8 => int8Array2Data(inArray)
        where(arrayDataInt8 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER16) then
        arrayDataInt16 => int16Array2Data(inArray)
        where(arrayDataInt16 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. INTEGER32) then
        arrayDataInt32 => int32Array2Data(inArray)
        where(arrayDataInt32 > 0) fitsArray = .true.
      else if (dataType(inArray) .eq. REAL32) then
        arrayDataReal32 => real32Array2Data(inArray)
        where(arrayDataReal32 > 0) fitsArray = .true.
!      else if (dataType(inArray) .eq. BOOLEAN) then
!        arrayDataBoolean => boolArray2Data(inArray)
!        fitsArray = arrayDataBoolean
      else ! this gives a DAL error if data type is none of the above.
        arrayDataReal64 => real64Array2Data(inArray)
        where(arrayDataReal64 > 0) fitsArray = .true.
      end if

    case(3)
!      call loadFitsArray_3d_wa_r32(inArray, cube)!!!, dataType)
!      arrayShape(:3) = shape(cube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(cube, imageShape)
call error('notSupported', '')

    case(4)
!      call loadFitsArray_4d_wa_r32(inArray, hyperCube)!!!, dataType)
!      arrayShape = shape(hyperCube)
!if (count(arrayShape .gt. 1) .gt. 2) call error('badHyperCubeShape','')
!      biggestDimensions(1,:) = maxloc(arrayShape)
!      mask(biggestDimensions(1,1)) = .false.
!      biggestDimensions(2,:) = maxloc(arrayShape, mask)
!      imageShape(1) = size(cube, biggestDimensions(1,1))
!      imageShape(2) = size(cube, biggestDimensions(2,1))
!      allocate(fitsArray(imageShape(1), imageShape(2)))
!      fitsArray = reshape(hyperCube, imageShape)
call error('notSupported', '')

    case default
call error('arrayRankTooBig', '')
    end select

  end subroutine loadFitsArray2dBoolean

  !---------------------------------------------------------------------
  subroutine mergeTables(inSetNames, inTabNames, outSetName&
  , outTabName, inColNames, indexColName, indexFileName, status)
  !---------------------------------------------------------------------
  ! Merges up to 10 tables into 1. At present the input tables must have the same number of rows.

    implicit none

    character(*), intent(in)            :: inSetNames(:),&
                                           inTabNames(size(inSetNames))
    character(*), intent(in)            :: outSetName
    character(*), intent(in),  optional :: outTabName
    character(*), intent(in),  optional :: indexColName,&
                                           indexFileName,&
                                           inColNames(:)
    integer,      intent(out), optional :: status

    character(255) ::&
      localOutTabName
    character(10) ::&
      setNumStr
    integer ::&
      numSets,&
      maxNumSets,&
      maxNumCols,&
      numRowsInFirstTab,&
      localStatus,&
      colNum,&
      setNum,&
      numColsTakenFromThisSet,&
      startSet
    integer, parameter ::&
      numColNameChars = 20,&
      numDigits = 1
    character(numColNameChars - (1 + numDigits)), allocatable ::&
      maxListOfColNames(:)
    ! The funny length of these strings is because we are going to construct the column names in the outset by adding an underscore plus the set number to the input column names, and we want output names that are (i) less than numColNameChars long and (ii) unique within the table.
    character(numColNameChars) ::&
      outColName
    logical, allocatable ::&
      columnIsPresentInAll(:)
    type(DataSetT) ::&
      firstSet,&
      set,&
      outSet
    type(TableT) ::&
      firstTab,&
      tab,&
      outTab
    integer, allocatable ::&
      indices(:),&
      allIndices(:,:)
integer, pointer :: xi(:)

    !-------------------------------------------------------------------
    ! Start

print *, '000'

    localStatus = 0
    numSets = size(inSetNames)
    if (numSets .le. 1) then
      if (present(status)) status = -1
  return
    end if
!    maxNumSets = 10**numDigits
maxNumSets = 10
    if (numSets .gt. maxNumSets) then
      if (present(status)) status = -2
  return
    end if

print *, '111'
    if (present(outTabName)) then
      localOutTabName = outTabName
    else
      localOutTabName = 'COMBINED'
    end if

    firstSet = dataSet(inSetNames(1), READ)
    firstTab = table(firstSet, inTabNames(1))
    numRowsInFirstTab = numberOfRows(firstTab)

    if (present(inColNames)) then
print *, '222'
      maxNumCols = size(inColNames)
      allocate(maxListOfColNames(maxNumCols)&
      , columnIsPresentInAll(maxNumCols))
  
      columnIsPresentInAll = .true. ! default
      maxListOfColNames    = inColNames
      startSet = 1

    else
      maxNumCols = numberOfColumns(firstTab)
      allocate(maxListOfColNames(maxNumCols)&
      , columnIsPresentInAll(maxNumCols))

      columnIsPresentInAll = .true. ! default
      maxListOfColNames   = ''     ! default

      do colNum = 1, maxNumCols
        if (len(name(column(firstTab, colNum - 1))) .gt. numColNameChars) then
          localStatus = localStatus + 1
          columnIsPresentInAll(colNum) = .false.
        else
          maxListOfColNames(colNum) = name(column(firstTab, colNum - 1))
        end if
      end do
      startSet = 2
    end if
    call release(firstSet)
print *, '333'

    do setNum = startSet, numSets
!print *, 'setNum, inSetNames(setNum), inTabNames(setNum): ', setNum, trim(inSetNames(setNum)), trim(inTabNames(setNum))
      set = dataSet(inSetNames(setNum), READ)
!print *, '  aaa'
      tab = table(set, inTabNames(setNum))
!print *, '  bbb'

      do colNum = 1, maxNumCols
!print *, '    ---'
!print *, 'colNum, maxListOfColNames(colNum): ', colNum, maxListOfColNames(colNum)
      if (.not. columnIsPresentInAll(colNum)) cycle
!print *, '    --+'
        if (.not. hasColumn(tab, maxListOfColNames(colNum)))&
          columnIsPresentInAll(colNum) = .false.
!print *, '    column ', maxListOfColNames(colNum), ' IsPresent = ', columnIsPresentInAll(colNum)
      end do

!print *, '  ccc'
      call release(set)
!print *, '  ddd'
    end do
print *, '444'

    if (.not. any(columnIsPresentInAll)) then
      if (present(status)) status = -3
  return
    end if
print *, '555'

    allocate(indices(numRowsInFirstTab))
print *, '666'
    if (.not. present(indexColName) .and. present(indexFileName)) then
print *, '777'
      allocate(allIndices(numRowsInFirstTab, numSets))
print *, '888'
      call loadIndices(indexFileName, allIndices, localStatus)
print *, '999'
      if (localStatus .lt. 0) then
print *, 'oooo! status=', status
        if (present(status)) status = localStatus
  return
      end if
print *, '000'

    else
!      indices = indicize(numRowsInFirstTab) - 1
xi => indicize(numRowsInFirstTab)
indices = xi - 1
deallocate(xi)
    end if

    outSet = dataSet(outSetName, CREATE)
print *, '111 localOutTabName=', '*'//trim(localOutTabName)//'*'
    outTab = addTable(outSet, trim(localOutTabName), numRowsInFirstTab)
print *, '222'
    do setNum = 1, numSets
print *, ''
print *, '  aaa'
      write(setNumStr, '(i10.10)') setNum-1
print *, '  bbb'
      set = dataSet(inSetNames(setNum), READ)
print *, '  ccc'
      tab = table(set, inTabNames(setNum))
print *, '  ddd'
      if (numberOfRows(tab) .ne. numRowsInFirstTab) then
        if (present(status)) status = -4
print *, '  oooo! numberOfRows(tab) (',numberOfRows(tab),') .ne. &
&numRowsInFirstTab (',numRowsInFirstTab,'). status=',status
  return
      end if
print *, '  eee'

      if (present(indexColName)) then
        call readColData(tab, indexColName, indices)
      else if (present(indexFileName)) then
        indices = allIndices(:, setNum)
      end if
print *, '  fff'

      numColsTakenFromThisSet = 0
print *, '  ggg'
      do colNum = 1, maxNumCols
      if (.not. columnIsPresentInAll(colNum)) cycle
        numColsTakenFromThisSet = numColsTakenFromThisSet + 1
        outColName = trim(maxListOfColNames(colNum))//'_'&
        //setNumStr(len(setNumStr) + 1 - numDigits:)
        call copyColumnInCorrectOrder(tab, maxListOfColNames(colNum)&
        , indices, outTab, outColName)
      end do

print *, '  hhh'
      if (numColsTakenFromThisSet .le. 0) localStatus = localStatus + 2

print *, '  iii'
      call release(set)
    end do
print *, '333'
    call release(outSet)

    if (present(status)) status = localStatus
print *, 'eeee nnnn dddd'
deallocate(indices)

  end subroutine mergeTables

  !---------------------------------------------------------------------
  subroutine loadIndices(indexFileName, allIndices, status)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: indexFileName
    integer, intent(out) :: allIndices(:,:), status

    integer :: ioUnit, numRows, numSets, lineNum, readStatus
    logical :: fileExists, fileOpened
    character(255) :: buffer(size(allIndices, 1))

    !-------------------------------------------------------------------
    ! Start

!print *, '  start of loadIndices.'
    status = 0

    numRows = size(allIndices, 1)
    numSets = size(allIndices, 2)
    inquire(file=indexFileName, exist=fileExists, opened=fileOpened)
    if (.not. fileExists) then
      status = -5
  return
    else if (fileOpened) then
      status = -6
  return
    end if

!    call get_free_io_unit(ioUnit)
call getFreeIoUnit(ioUnit)
    open(ioUnit, file=indexFileName, status='old', form='formatted'&
    , action='read', pad='yes')
      lineNum = 0
      do
        lineNum = lineNum + 1
!print *, ''
!print *, '    lineNum=', lineNum
        read(ioUnit, '(a255)', iostat=readStatus) buffer(lineNum)
!        read(ioUnit, *, iostat=readStatus) buffer(lineNum)
      if (readStatus .ne. 0) exit
!print *, lineNum, '*'//buffer(lineNum)//'* **'//trim(buffer(lineNum))//'**'
        call pickBufferApart(trim(buffer(lineNum)), allIndices(lineNum,:), status)
!print *, '    pickBufferApart status=', status
!print *, '    1st index=', allIndices(lineNum,1)
!print *, '    2nd index=', allIndices(lineNum,2)
      if (status .lt. 0) exit
      end do    
    close(ioUnit)

!print *, ''
!print *, '  end of loadIndices. status=', status

  end subroutine loadIndices

  !---------------------------------------------------------------------
  subroutine pickBufferApart(buffer, indices, status)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: buffer
    integer, intent(out) :: indices(:), status

    integer :: tabPosition, nonTabStart, nonTabEnd, numSets, setNum

    !-------------------------------------------------------------------
    ! Start

!print *, 'start of pickBufferApart'

    status = 0

!print *, '**'//trim(buffer)//'**'
    numSets = size(indices)

    if (len(buffer) .le. 0) then
      status = -1
  return
    end if
    tabPosition = index(buffer, achar(9))
    if (tabPosition .le. 0) then ! no tabs found.
      status = -2
  return
    end if
    if (tabPosition .eq. 1) then ! the buffer starts with a tab. Not allowed.
      status = -3
  return
    end if

    nonTabStart = 1 ! hopefully it is the start of an integer.
    setNum = 0
!print *, '      len(buffer):', len(buffer)
!print *, '      setNum:', setNum
    do
!print *, '      '
!print *, '      Start of do.'
      tabPosition = index(buffer(nonTabStart:), achar(9))
!print *, '      tabPosition:', tabPosition
      if (tabPosition .eq. len(buffer)) then ! not allowed.
        status = -4
!print *, '      ooo status -4, tabPosition .eq. len(buffer).'
  return
      else if (tabPosition .le. 0) then ! no further tabs found.
        nonTabEnd = len(buffer)
!print *, '      tabPosition .le. 0. nonTabEnd:', nonTabEnd
      else
        nonTabEnd = tabPosition - 1
!print *, '      tabPosition .gt. 0. nonTabEnd:', nonTabEnd
      end if

!print *, '      nonTabStart .gt. nonTabEnd? exit loop if so:', (nonTabStart .gt. nonTabEnd)
    if (nonTabStart .gt. nonTabEnd) exit ! because there are no more integers.

      setNum = setNum + 1
!print *, '      setNum, numSets:', setNum, numSets
      if (setNum .gt. numSets) then
        status = -5
!print *, '      ooo status -5, setNum .gt. numSets.'
  return
      end if

!print *, '      just before call to strToInt. buffer(nonTabStart:nonTabEnd), ie buffer('&
!,nonTabStart,':',nonTabEnd,'), is ', buffer(nonTabStart:nonTabEnd)
      call strToInt(buffer(nonTabStart:nonTabEnd), indices(setNum), status)
!print *, '      indices(setNum), status:', indices(setNum), status
  if (status .lt. 0) return
      indices(setNum) = indices(setNum) - 1

    if (tabPosition .le. 0 .or. nonTabEnd .ge. len(buffer)) exit
      nonTabStart = tabPosition + 1
!print *, '      nonTabStart:', nonTabStart
    end do

    if (setNum .lt. numSets) then
      status = -6
  return
    end if

  end subroutine pickBufferApart

  !---------------------------------------------------------------------
  subroutine copyColumnInCorrectOrder(inTab, inColName&
  , indices, outTab, outColName)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: inTab, outTab
    character(*), intent(in) :: inColName, outColName
    integer(int32), intent(in) :: indices(:)

    type(ColumnT) :: inCol, outCol
    integer(int8),  dimension(:), pointer ::&
      inColDataInt8, outColDataInt8
    integer(int16), dimension(:), pointer ::&
      inColDataInt16, outColDataInt16
    integer(int32), dimension(:), pointer ::&
      inColDataInt32, outColDataInt32
    real(single),   dimension(:), pointer ::&
      inColDataReal32, outColDataReal32
    real(double),   dimension(:), pointer ::&
      inColDataReal64, outColDataReal64

    !-------------------------------------------------------------------
    ! Start

    inCol  = column( inTab,  inColName)

    select case(dataType(inCol))
    case(INTEGER8)
      outCol = addColumn(outTab, outColName, INTEGER8)
      inColDataInt8    => int8Data(inCol)
      outColDataInt8   => int8Data(outCol)
      outColDataInt8   = inColDataInt8(indices)
    case(INTEGER16)
      outCol = addColumn(outTab, outColName, INTEGER16)
      inColDataInt16   => int16Data(inCol)
      outColDataInt16  => int16Data(outCol)
      outColDataInt16  = inColDataInt16(indices)
    case(INTEGER32)
      outCol = addColumn(outTab, outColName, INTEGER32)
      inColDataInt32   => int32Data(inCol)
      outColDataInt32  => int32Data(outCol)
      outColDataInt32  = inColDataInt32(indices)
    case(REAL32)
      outCol = addColumn(outTab, outColName, REAL32)
      inColDataReal32  => real32Data(inCol)
      outColDataReal32 => real32Data(outCol)
      outColDataReal32 = inColDataReal32(indices)
    case default ! this gives a DAL error if data type is not REAL64 either:
      outCol = addColumn(outTab, outColName, REAL64)
      inColDataReal64  => real64Data(inCol)
      outColDataReal64 => real64Data(outCol)
      outColDataReal64 = inColDataReal64(indices)
    end select

    call release(inCol)
    call release(outCol)

  end subroutine copyColumnInCorrectOrder

  !---------------------------------------------------------------------
  subroutine saveOutImage(newImage, imageName, newImageName&
  , xsize, ysize, xStart, yStart)
  !---------------------------------------------------------------------
  ! This saves newImage to a dataset and copies over the wcs and ogip keywords from imageName to newImageName. Hence it can be used to output an image after it has been processed in some way.
  !
  ! The optional xsize, ysize, xStart and yStart arguments are to cater for the situation in which the original image was surrounded by a rectangular border of zero values which it is not desired to process. In some tasks (eg asmooth) this blank border is trimmed off before the smoothing is done. However it is desirable that the processed image be the same size as the input, so this blank border (if any) can be reconstituted via these arguments. The pair (xsize, ysize) gives the size of the original image, including blank border; (xStart, yStart) gives the column and row number respectively of the first non-all-blank column or row. The array newImage() is assumed to start at this point.

    implicit none

    real(single), intent(in) ::&
      newImage(:,:)
    character(*), intent(in) ::&
      imageName, newImageName
    integer(int16), intent(in), optional :: xsize, ysize, xStart, yStart

    type(DataSetT) :: inSet, outSet
    type(ArrayT) :: inArray, outArray
    real(single), dimension(:,:), pointer ::&
      arrayDataReal32
    integer(int32), dimension(:), pointer ::&
      dimVector
    integer(int16) ::&
      xsizeLocal, ysizeLocal, xStartLocal, yStartLocal,&
      xFinis, yFinis

    !-------------------------------------------------------------------
    ! Start:

    if (present(xsize)) then
      xsizeLocal = xsize
    else
      xsizeLocal = size(newImage, 1)
    end if
    if (present(ysize)) then
      ysizeLocal = ysize
    else
      ysizeLocal = size(newImage, 2)
    end if
    if (present(xStart)) then
      xStartLocal = xStart
    else
      xStartLocal = 1
    end if
    if (present(yStart)) then
      yStartLocal = yStart
    else
      yStartLocal = 1
    end if

    inSet = dataSet(imageName, READ)
    inArray = array(inSet, 0, READ)
    outSet = dataSet(newImageName, CREATE)
    allocate(dimVector(0:1))
    dimVector(0) = xsizeLocal
    dimVector(1) = ysizeLocal
    xFinis = xStartLocal + size(newImage, 1) - 1
    yFinis = yStartLocal + size(newImage, 2) - 1
    outArray = addArray(outSet, 'PRIMARY', REAL32, dimVector, position=0)
    call setAttributes(attributable(outSet), attributable(inSet))

    call copyVitalArrayKwds(inArray, outArray)
    ! This is done instead of a total copy of all array-header keywords because
    ! doing the latter stuffs up the DSS.

    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = 0.0
    arrayDataReal32(xStartLocal-1:xFinis-1, yStartLocal-1:yFinis-1) = newImage

    call release(outSet)
    call release(inSet)

    deallocate(dimVector)

  end subroutine saveOutImage

  !---------------------------------------------------------------------
  subroutine copyVitalArrayKwds(inArray, outArray)
  !---------------------------------------------------------------------

    implicit none

    type(ArrayT), intent(in) :: inArray, outArray

    integer :: status

    !-------------------------------------------------------------------
    ! Start

    call copyAttribute(inArray, 'SLCTEXPR', outArray, status)
    call copyAttribute(inArray, 'HDUCLASS', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS1', outArray, status)
    call copyAttribute(inArray, 'HDUCLAS2', outArray, status)
    call copyAttribute(inArray, 'HDUVERS1', outArray, status)

    call copyAttribute(inArray, 'CTYPE1', outArray, status)
    call copyAttribute(inArray, 'CRPIX1', outArray, status)
    call copyAttribute(inArray, 'CRVAL1', outArray, status)
    call copyAttribute(inArray, 'CUNIT1', outArray, status)
    call copyAttribute(inArray, 'CDELT1', outArray, status)
    call copyAttribute(inArray, 'CTYPE2', outArray, status)
    call copyAttribute(inArray, 'CRPIX2', outArray, status)
    call copyAttribute(inArray, 'CRVAL2', outArray, status)
    call copyAttribute(inArray, 'CUNIT2', outArray, status)
    call copyAttribute(inArray, 'CDELT2', outArray, status)

    call copyAttribute(inArray, 'EXPOSURE', outArray, status)

  end subroutine copyVitalArrayKwds

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module fits_utils
