!+
! Module: linear
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines to do with least-squares estimation and solution of linear equations.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module linear

use dal
!*** change to non-dal forms.
!  use param  

  implicit none 

  private 

  public solveLinearEquations

contains 

  !---------------------------------------------------------------------
  subroutine solveLinearEquations(m, x, y, pivotPlease, status)
  !---------------------------------------------------------------------
  ! This sub solves the matrix equation m * x = y for matrix m and vector
  ! x and y. It makes use of the fact that a (non-singular?) matrix can be
  ! decomposed into the product of two matrices L and U, L being a lower-
  ! triangular matrix with 1s on the diagonal, U being an upper-triangular
  ! matrix. The fact that L has 1s on its diagonal permits both matrices to
  ! be stored in the same array, as in done here. Partial pivoting is done
  ! if the pivotPlease flag is set .true..
  !
  ! This sub has been tested pretty well.

    implicit none

    real(double),    intent(in)            :: m(:,:), y(:)
    logical, intent(in),  optional :: pivotPlease
    real(double),    intent(out)           :: x(:)
    integer, intent(out), optional :: status

    real(double), allocatable :: lu(:,:), tempY(:)
    integer :: numElements, row, col, r, c, maxrow, cc, localStatus
    real(double) :: max, temp
    logical :: localPivotPlease

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0

    x = 0.0

    numElements = size(y)
    if (size(x) .ne. numElements) then
      if (present(status)) status = -1
  return
    end if
    if (size(m, 1) .ne. numElements) then
      if (present(status)) status = -2
  return
    end if
    if (size(m, 2) .ne. numElements) then
      if (present(status)) status = -3
  return
    end if
    if (numElements .lt. 1) then
      if (present(status)) status = -4
  return
    end if

    if (present(pivotPlease)) then
      localPivotPlease = pivotPlease
    else
      localPivotPlease = .false.
    end if

    allocate(lu(numElements, numElements), tempY(numElements))
    lu = transpose(m)
    tempY = y

    ! LU factorisation:
    do col = 1, numElements - 1
      if (localPivotPlease) then ! partial pivot.
        max = abs(lu(col, col))
        maxrow = col
        do r = col + 1, numElements
          if (abs(lu(r, col)) .gt. max) then
            max = abs(lu(r, col))
            maxrow = r
          end if
        end do !r
!*** can do this more efficiently with f95 inbuilt functions.

        do c = 1, numElements
          temp = lu(col, c)
          lu(col, c) = lu(maxrow, c)
          lu(maxrow, c) = temp
        end do !c
        temp = tempY(col)
        tempY(col) = tempY(maxrow)
        tempY(maxrow) = temp
      end if ! pivotPlease

!print *, 'col', col
      do row = col + 1, numElements
        if (lu(col, col) .eq. 0.0) then
          if (present(status)) status = -5
          deallocate(lu, tempY)
  return
        end if

!print *, 'row', row
        lu(row, col) = lu(row, col) / lu(col, col)
        do cc = col + 1, numElements
          lu(row, cc) = lu(row, cc) - lu(row, col) * lu(col, cc)
        end do !cc
      end do !row
    end do !col

    ! Forward substitution to solve L(Ux) = tempY (leaving Ux in array x):
    x(1) = tempY(1)
    do row = 2, numElements
      x(row) = tempY(row)
      do col = 1, row - 1
        x(row) = x(row) - lu(row, col) * x(col)
      end do !col
    end do !row

    ! Back substitution to solve Ux = tempY:
    if (lu(numElements, numElements) .eq. 0.0) then
      if (present(status)) status = -6
      deallocate(lu, tempY)
  return
    end if
    x(numElements) = x(numElements) / lu(numElements, numElements)

    do row = numElements - 1, 1, -1
      do col = row + 1, numElements
        x(row) = x(row) - lu(row, col) * x(col)
      end do !col

      if (lu(row, row) .eq. 0.0) then
        if (present(status)) status = -6
        deallocate(lu, tempY)
  return
      end if
      x(row) = x(row) / lu(row, row)
    end do !row

    deallocate(lu, tempY)

  end subroutine solveLinearEquations

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module linear
