!+
! Module: mathutils
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some assorted routines.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module mathutils

  use dal
  use param  
  use errorHandling

  use constants

  implicit none 

  private 

  public pointInEllipseWithPhase
  public calcEllipseFromPhase
  public ellipseAngleToPhase
  public vectorCross
  public hybridNewtonsMethod

  interface ellipseAngleToPhase
    module procedure ellipseAngleToPhaseScalar
    module procedure ellipseAngleToPhaseVector
  end interface

  interface vectorCross
    module procedure vectorCrossReal
    module procedure vectorCrossDble
  end interface

  type, public :: statusType
    integer :: status, maxNeg, numPosBits
  end type

!logical :: testFlag = .false.

contains 

  !---------------------------------------------------------------------
  function pointInEllipseWithPhase(x, y, xAmp, yAmp, phase)
  !---------------------------------------------------------------------
  ! The ellipse is assumed to be centred about the origin. It is defined
  ! by two amplitudes and a phase.
  !
  ! The function returns -1 if the point (x, y) lies fully within the
  ! ellipse; 0 if it is on the border; and 1 if it is fully outside the
  ! ellipse.

    implicit none

    real, intent(in) ::&
      x,&
      y,&
      xAmp,&
      yAmp,&
      phase

    integer :: pointInEllipseWithPhase
    real :: relX, termA, termB, topY, botY

    !-------------------------------------------------------------------
    ! Start

    relX = x / xAmp

    if (abs(relX) > 1.0) then
      pointInEllipseWithPhase = 1
  return
    else if (abs(relX) == 1.0) then
      pointInEllipseWithPhase = 0
  return
    end if

    termA = relX * sin(phase)
    termB = sqrt(1 - relX * relX) * cos(phase)
    topY = yAmp * (termA + termB)
    botY = yAmp * (termA - termB)

    if (y > topY .or. y < botY) then
      pointInEllipseWithPhase = 1
    else if (y == topY .or. y == botY) then
      pointInEllipseWithPhase = 0
    else
      pointInEllipseWithPhase = -1
    end if

  end function pointInEllipseWithPhase

  !---------------------------------------------------------------------
  subroutine calcEllipseFromPhase(xAmp, yAmp, phase, xVals, yVals)
  !---------------------------------------------------------------------
  ! This sub calculates coordinates of samples of an ellipse. The larger
  ! the size of the vectors xVals and yVals, the more accurately is the
  ! ellipse sampled. The intention is, after scaling and addition of
  ! offsets, to send the resulting polynomial to the pgplot routine pgpoly,
  ! viz:
  !
  !   call pgpoly(size(xVals), xVals, yVals)
  !

    implicit none 

    real,  intent(in) :: xAmp,&
                         yAmp,&
                         phase
    real, intent(out) :: xVals(:),&
                         yVals(size(xVals))

    integer :: point, numPoints
    real :: pointAngle

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(xVals)

    xVals = 0.0
    yVals = 0.0

    do point = 1, numPoints
      pointAngle = (point / real(numPoints)) * TwoPi
      xVals(point) = xAmp * cos(pointAngle)
      yVals(point) = yAmp * sin(pointAngle + phase)
    end do

  end subroutine calcEllipseFromPhase

  !---------------------------------------------------------------------
  subroutine ellipseAngleToPhaseScalar(xSemi, ySemi, angleDeg, xAmp, yAmp&
  , phase)
  !---------------------------------------------------------------------
  ! Converts between the two ellipse representations.

    implicit none 

    real,  intent(in)  :: xSemi,&
                          ySemi,&
                          angleDeg
    real,  intent(out) :: xAmp,&
                          yAmp,&
                          phase

    real  ::&
      xSemiVector(1), ySemiVector(1), angleDegVector(1),&
      xAmpVector(1), yAmpVector(1), phaseVector(1)

    !-------------------------------------------------------------------
    ! Start:

    xSemiVector(1)    = xSemi
    ySemiVector(1)    = ySemi
    angleDegVector(1) = angleDeg

    call ellipseAngleToPhaseVector(xSemiVector, ySemiVector, angleDegVector&
      , xAmpVector, yAmpVector, phaseVector)

    xAmp  = xAmpVector(1)
    yAmp  = yAmpVector(1)
    phase = phaseVector(1)

  end subroutine ellipseAngleToPhaseScalar

  !---------------------------------------------------------------------
  subroutine ellipseAngleToPhaseVector(xSemi, ySemi, angleDeg, xAmp, yAmp&
  , phase)
  !---------------------------------------------------------------------
  ! Converts between the two ellipse representations.

    implicit none 

    real,  intent(in)  :: xSemi(:),&
                          ySemi(size(xSemi)),&
                          angleDeg(size(xSemi))
    real,  intent(out) :: xAmp(size(xSemi)),&
                          yAmp(size(xSemi)),&
                          phase(size(xSemi))

    real :: cosAngle, sinAngle, x, y
    integer :: i
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    do i = 1, size(xSemi)
      cosAngle = cos(angleDeg(i) * PiOn180)
      sinAngle = sin(angleDeg(i) * PiOn180)
      xAmp(i) = sqrt(xSemi(i) * xSemi(i) * cosAngle * cosAngle&
                + ySemi(i) * ySemi(i) * sinAngle * sinAngle)

      yAmp(i) = sqrt(xSemi(i) * xSemi(i) * sinAngle * sinAngle&
                + ySemi(i) * ySemi(i) * cosAngle * cosAngle)

      if (ySemi(i) .ne. 0.0 .or. xSemi(i) .ne. 0.0) then
y = (xSemi(i) * xSemi(i) - ySemi(i) * ySemi(i)) * sinAngle * cosAngle
x = ySemi(i) * xSemi(i)
        phase(i) = atan2(y, x)
if (testFlag) print *, 'i, x, y, phase(i): ', i, x, y, phase(i)

      else
        phase(i) = PiOverTwo
      end if
    end do

  end subroutine ellipseAngleToPhaseVector

  !---------------------------------------------------------------------
  function vectorCrossReal(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real, dimension(3), intent(in) :: vectorA, vectorB
    real :: vectorCrossReal(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossReal(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossReal(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossReal(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossReal

  !---------------------------------------------------------------------
  function vectorCrossDble(vectorA, vectorB)
  !---------------------------------------------------------------------

    implicit none

    real(kind(1d0)), dimension(3), intent(in) :: vectorA, vectorB
    real(kind(1d0)) :: vectorCrossDble(3)

    !-------------------------------------------------------------------
    ! Start

    vectorCrossDble(1) = vectorA(2) * vectorB(3) - vectorA(3) * vectorB(2)
    vectorCrossDble(2) = vectorA(3) * vectorB(1) - vectorA(1) * vectorB(3)
    vectorCrossDble(3) = vectorA(1) * vectorB(2) - vectorA(2) * vectorB(1)

  end function vectorCrossDble

  !---------------------------------------------------------------------
  subroutine hybridNewtonsMethod(startX0, startX1, y, dy, solutionX&
  , maxNumIterations, convergenceCriterion, stat, status)
  !---------------------------------------------------------------------
  ! Adapted from the hybrid bisection/Newton-Raphson method in Numerical
  ! Methods.
  !
  ! This subroutine solves the equation y(x) = 0 by Newton's Method. The
  ! user must construct two functions in the calling module, one of which
  ! returns y and the other dy/dx for a given x. These functions (and their
  ! single argument x) must be of type real. The user should pass their
  ! names in the call to hybridNewtonsMethod.
  !
  ! The starting values startX0 and startX1 should bracket the zero point
  ! but it is not necessary that they be in any order.

    implicit none

    real,    intent(in)            :: startX0, startX1
    real,    intent(out)           :: solutionX
    integer, intent(in),  optional :: maxNumIterations
    real,    intent(in),  optional :: convergenceCriterion
    integer, intent(out), optional :: status
    type(statusType), intent(out), optional :: stat

    interface
      real function y(x)
        real, intent(in) :: x
      end function y
      real function dy(x)
        real, intent(in) :: x
      end function dy
    end interface

    real :: yValue, yDeriv, localConvergenceCriterion, startY0, startY1,&
      xLo, xHi, dx, dxOld, oldSolutionX
    integer :: localStatus, localMaxNumIter, j
    type(statusType) :: localStat

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localStat%status = 0 ! default
    localStat%numPosBits = 1
    localStat%maxNeg = -2
    if (present(stat)) stat = localStat

    solutionX = 0.0 ! Don't like to leave it with no value, even if the routine
    ! can't calculate it.

    if (present(maxNumIterations)) then
      localMaxNumIter = maxNumIterations
    else
      localMaxNumIter = 100
!*** can do better than this.
    end if

    startY0 = y(startX0)
    startY1 = y(startX1)

    if (startY0 .eq. 0.0) then
      solutionX = startX0
  return
    end if
    if (startY1 .eq. 0.0) then
      solutionX = startX1
  return
    end if

    if (startX0 .eq. startX1) then
      if (present(status)) status = -1 ! no space between bounds.
      if (present(stat)) stat%status = -1 ! no space between bounds.
  return
    end if

    if (startY0 .gt. 0.0 .and. startY1 .gt. 0.0&
    .or. startY0 .lt. 0.0 .and. startY1 .lt. 0.0) then
      if (present(status)) status = -2 ! root must lie between startX0 and
      if (present(stat)) stat%status = -2 ! root must lie between startX0 and
      ! startX1.
  return
    end if

    ! Orient bounds such that y(xLo) < 0:
    if (startY0 .lt. 0.0) then
      xLo = startX0
      xHi = startX1
    else
      xLo = startX1
      xHi = startX0
    end if

    solutionX = 0.5 * (xLo + xHi)
    dxOld = xHi - xLo
    dx = dxOld

    if (present(convergenceCriterion)) then
      localConvergenceCriterion = abs(convergenceCriterion)
    else
      localConvergenceCriterion = 0.5 * abs(xLo + xHi)&
      * (10.0 ** (1-precision(dx)))
    end if

    yValue = y(solutionX)
    yDeriv = dy(solutionX)
    do j = 1, localMaxNumIter
      dxOld = dx

      if (((solutionX - xHi) * yDeriv - yValue)&
      * ((solutionX - xLo) * yDeriv - yValue) .gt. 0.0&
      .or. abs(2.0 * yValue) .gt. abs(dxOld * yDeriv)) then
        dx = 0.5 * (xHi - xLo)
        solutionX = xLo + dx
    if (solutionX .eq. xLo) exit

      else
        dx = yValue / yDeriv
        oldSolutionX = solutionX
        solutionX = solutionX - dx
    if (solutionX .eq. oldSolutionX) exit
      end if

    if (abs(dx) .lt. localConvergenceCriterion) exit

      yValue = y(solutionX)
      yDeriv = dy(solutionX)

      if (yValue .lt. 0.0) then
        xLo = solutionX
      else
        xHi = solutionX
      end if
    end do
    if (j .ge. localMaxNumIter) then
      localStatus = 1
      localStat%status = 1
    end if

    if (present(status)) status = localStatus
    if (present(stat)) stat = localStat

  end subroutine hybridNewtonsMethod

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module mathutils
